# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging
import os
import pygame

import trosnoth.data.music as music
from trosnoth.data import getPath

log = logging.getLogger(__name__)

# NO_MORE_MUSIC must not collide with any pygame events.
NO_MORE_MUSIC = 234

pygame.mixer.music.set_endevent(NO_MORE_MUSIC)

class MusicManager(object):
    '''Manages the music.'''

    def __init__(self):
        try:
            pygame.mixer.init(44100)
        except:
            log.error('Could not initialise audio.')
        self.index = 0
        self.volume = 100
        self.filenames = []
        for f in os.listdir(getPath(music)):
            if f.endswith('.ogg'):
                self.addMusicFile(f)

    def processEvent(self, event):
        if event.type == NO_MORE_MUSIC:
            self.musicEnded()
            return None
        else:
            return event

    def addMusicFile(self, filename):
        self.filenames.append(filename)

    def playMusic(self):
        if not pygame.mixer.get_init():
            return

        curFile = self.filenames[self.index]
        pygame.mixer.music.load(getPath(music, curFile))
        pygame.mixer.music.play(0)
        self.setVolume(self.volume)

        # Queue one more.
        self.queueNext()

    def queueNext(self):
        # Queue the next one.
        self.index = (self.index + 1) % len(self.filenames)
        curFile = self.filenames[self.index]
        pygame.mixer.music.queue(getPath(music, curFile))

    def musicEnded(self):
        self.queueNext()
        self.setVolume(self.volume)

    def stopMusic(self):
        if pygame.mixer.get_init():
            pygame.mixer.music.stop()

    def isMusicPlaying(self):
        if not pygame.mixer.get_init():
            return False

        return pygame.mixer.music.get_busy()

    def setVolume(self, volume):
        if not pygame.mixer.get_init():
            return

        self.volume = volume
        # Valid range is 0.0 - 1.0
        decimalVolume = volume / 100.0
        pygame.mixer.music.set_volume(decimalVolume)




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import json
import os
import operator

from trosnoth.data import getPath, makeDirs
from trosnoth.utils.utils import hasher, stripPunctuation
from trosnoth.data import statGeneration
from trosnoth.gamerecording.achievementlist import (leaderAchievements,
        additionalLeaders)

def generateHtml(htmlPath, statPath):

    def plural(value):
        if value == 1:
            return ""
        else:
            return "s"

    def add(value, statName=None, altText=None, spacing=True, className=None):
        if altText is not None:
            points = pointValues[altText] * value
            if type(points) == float:
                points = '%2.2f' % points
            altText = ' title="%s point%s\"' % (points, plural(points))
        else:
            altText = ''

        if className is not None:
            classText = ' class="%s"' % (className)
        else:
            classText = ''

        if type(value) == float:
            value = '%2.2f' % value

        if statName is None:
            html.append("\t\t\t\t<td%s%s>%s</td>" % (altText, classText, value))
        else:
            nbsp = ""
            if spacing is False:
                pluralStr = ""
            else:
                nbsp = "&nbsp;"
                pluralStr = plural(value)
            html.append("\t\t\t\t<td%s>%s%s%s%s</td>" % (altText, value, nbsp,
                    statName, pluralStr))

    def addList(title, data):
        html.append("\t\t\t\t\t<li><b>%s:</b> " % title)

        if len(data) == 0:
            html.append("\t\t\t\t\t\tNone")
        else:
            data = sorted(data.iteritems(), key=operator.itemgetter(1),
                    reverse=True)
            string = []
            for details in data:
                string.append("%s (%d)" % details)
            html.append("\t\t\t\t\t\t" + ", ".join(string))

        html.append("\t\t\t\t\t</li>")

    def accuracy(shotsHit, shotsFired):
        try:
            return ((shotsHit ** 2) / (shotsFired + 0.0)) * 30
        except ZeroDivisionError:
            return 0

    from trosnoth.gamerecording.gamerecorder import statDir

    makeDirs(statDir)

    if statPath == '':
        files = os.listdir(statDir)
    else:
        files = [statPath]

    playerStats = {}
    statNames = ['aliveStreak', 'deaths', 'killStreak', 'kills', 'roundsLost',
            'roundsWon', 'shotsFired', 'shotsHit', 'starsEarned', 'starsUsed',
            'starsWasted', 'tagStreak', 'timeAlive', 'timeDead', 'zoneAssists',
            'zoneTags']
    statEnemies = ['playerDeaths', 'playerKills', 'upgradesUsed']
    pointValues = {'kills': 10, 'deaths': 1, 'zoneTags': 20, 'accuracy': 20,
            'starsUsed': 3, 'zoneAssists': 5}
    leaders = map(str.lower, leaderAchievements.keys() +
            additionalLeaders.keys())

    tableHeaders = [
        ['#', 'Nick', 'Kills', 'Deaths', 'KDR', 'Zone Tags', 'Shots Fired',
            'Shots Hit', 'Accuracy', 'Stars Used', 'Killed the most:',
            'Died the most to:', 'Points'],
        ['#', 'Nick', 'Stars Earned', 'Stars Used', 'Stars Wasted',
                'Favourite Upgrade', 'Time Alive', 'Time Dead', 'ADR',
                'Longest Life', 'Points'],
        ['#', 'Nick', 'Kills', 'Kill Streak', 'Zone Tags', 'Zone Assists',
                'Tag Streak', 'Points'],
    ]

    tableNames = ['General Overview', 'Stars and Time', 'Kills and Tags']

    for x in range(0, len(tableNames)):
        style = ""
        if x == 0:
            style = " style='color: black;'"
        tableNames[x] = ('<span class="name topLink" id="link%s" '
                'onClick="navigate(\'%s\', %d)"%s>%s</span>' %
                (x, x, len(tableHeaders[x]), style, tableNames[x]))

    navigation = " &ndash; ".join(tableNames)

    html = []
    fileMatrix = {}

    for filename in files:

        if filename[-9:] != ".trosstat":
            filename = filename + ".trosstat"
        statLocation = os.path.join(statPath, filename)

        try:
            statFile = open(statLocation)
        except IOError:
            raise Exception("'%s' does not exist!" % filename)

        loadedStats = json.load(statFile)

        for nick in loadedStats['players']:
            if nick not in playerStats:
                playerStats[nick] = loadedStats['players'][nick]
                fileMatrix[nick] = [filename]
            else:
                for stat in statNames:
                    playerStats[nick][stat] += (
                            loadedStats['players'][nick][stat])
                for stat in statEnemies:
                    for enemy in loadedStats['players'][nick][stat]:
                        if enemy not in playerStats[nick][stat]:
                            playerStats[nick][stat][enemy] = 0
                        playerStats[nick][stat][enemy] += (
                                loadedStats['players'][nick][stat][enemy])
                fileMatrix[nick].append(filename)

    ranking = {}
    allData = {}

    for nick in playerStats:
        data = playerStats[nick]
        try:
            data['accuracy'] = (100.0 * data['shotsHit']) / data['shotsFired']
        except ZeroDivisionError:
            data['accuracy'] = 0

        for stat in statEnemies:
            data[stat + "Full"] = data[stat].copy()
            highest = 0
            highestName = "----"
            names = data[stat]
            for k, v in names.items():
                if v > highest:
                    highest = v
                    highestName = k
            if highest == 0:
                data[stat] = highestName
            else:
                data[stat] = "%s (%s)" % (highestName, highest)

        data['score'] = 0
        for stat, value in pointValues.items():
            points = data[stat] * value
            data['score'] += points

        try:
            data['kdr'] = '%2.2f' % (float(data['kills']) / data['deaths'])
        except ZeroDivisionError:
            data['kdr'] = "----"

        try:
            data['adr'] = '%2.2f' % (float(data['timeAlive']) /
                    data['timeDead'])
        except ZeroDivisionError:
            data['adr'] = "----"

        ranking[nick] = data['score']
        allData[nick] = data

    rankingList = sorted(ranking.iteritems(), key=operator.itemgetter(1),
            reverse=True)
    ranking = {}

    rankCount = 0

    html.append("\t\t<table class='ladder'>");

    for count in range(0, len(tableNames)):
        style = ""
        if count != 0:
            style = " style='display: none;'"
        html.append("\t\t\t<tr class='allRows group%s'%s>" % (count, style))
        for caption in tableHeaders[count]:
            html.append("\t\t\t\t<th>%s</th>" % caption)
        html.append("\t\t\t</tr>")

    teamNames = ('Blue', 'Red')
    if loadedStats.has_key('winningTeamId'):
        winningTeamId = loadedStats['winningTeamId']
        if winningTeamId == 'A':
            winText = '%s Team won' % (teamNames[0],)
            colour = 'navy'
        elif winningTeamId == 'B':
            winText = '%s Team won' % (teamNames[1],)
            colour = 'maroon'
        else:
            winText = 'Game was a draw'
            colour = 'green'
    else:
        winText = 'Game was not finished'
        colour = 'gray'

    html.append('<p class="wintext" style="color: %s;">%s</p>' % (colour,
            winText))
    for pair in rankingList:

        nick = pair[0]
        rankCount += 1
        rankStr = str(rankCount)

        classy = ""
        if stripPunctuation(nick).lower() in leaders:
            classy = " leader";
            rankCount -= 1
            rankStr = '--'

        data = allData[nick]

        if data['bot']:
            classy = " bot";
            rankCount -= 1
            rankStr = 'B'

        nickHash = hasher(nick).hexdigest()

        for count in range(0, len(tableNames)):

            style = ""
            if count != 0:
                style = " style='display: none;'"

            html.append("\t\t\t<tr class='allRows group%s%s'%s>" % (count,
                    classy, style))
            if (data['team'] == 'A'):
                bgColour='blueteam'
            elif (data['team'] == 'B'):
                bgColour='redteam'
            elif (data['team'] == '\x00'):
                bgColour='rogue'
            add('<strong>%s</strong>' % rankStr, className=bgColour)
            add('<span class="name" onClick="toggle(\'details-%s\')">%s</span>'
                    % (nickHash, nick))

            if count == 0:
                add(data['kills'], 'kill', 'kills')
                add(data['deaths'], 'death', 'deaths')
                add(data['kdr'])
                add(data['zoneTags'], 'tag', 'zoneTags')
                add(data['shotsFired'], 'shot')
                add(data['shotsHit'], 'shot')
                add(data['accuracy'], '%', 'accuracy', False)
                add(data['starsUsed'], 'star', 'starsUsed')
                add(data['playerKills'])
                add(data['playerDeaths'])
            elif count == 1:
                add(data['starsEarned'], 'star')
                add(data['starsUsed'], 'star', 'starsUsed')
                add(data['starsWasted'], 'star')
                add(data['upgradesUsed'])
                add(int(data['timeAlive']), 'second')
                add(int(data['timeDead']), 'second')
                add(data['adr'])
                add(int(data['aliveStreak']), 'second')
            elif count == 2:
                add(data['kills'], 'kill', 'kills')
                add(data['killStreak'], 'kill')
                add(data['zoneTags'], 'tag', 'zoneTags')
                add(data['zoneAssists'], 'assist', 'zoneAssists')
                add(data['tagStreak'], 'zone')
            elif count == 3:
                add(data['shotsFired'], 'shot')
                add(data['shotsHit'], 'shot')
                add(data['accuracy'], '%', spacing = False)
                old = data['accuracy'] * 20
                new = accuracy(data['shotsHit'], data['shotsFired'])
                add(old, 'point')
                add(new, 'point')
                add(new - old)

            add('<strong>%2.2f</strong>' % data['score'])
            html.append("\t\t\t</tr>")

            if count == len(tableNames) - 1:
                html.append("\t\t\t<tr id='details-%s' style='display: none;'>"
                        % nickHash)
                html.append("\t\t\t\t<td colspan='%d' class='details'"
                        "style='text-align: left;'>" % len(tableHeaders[0]))
                html.append("\t\t\t\t\t<ul>")

                addList("Players killed", data['playerKillsFull'])
                addList("Players died to", data['playerDeathsFull'])
                addList("Upgrades used", data['upgradesUsedFull'])

                html.append("\t\t\t\t\t</ul>")
                html.append("\t\t\t\t</td>")
                html.append("\t\t\t</tr>")

    html.append("\t\t</table>")

    html = "\n" + "\n".join(html) + "\n"

    baseHTML = open(getPath(statGeneration, 'statGenerationBase.htm'),
            'r').read()

    html = baseHTML.replace("[[TABLE]]", html)
    html = html.replace("[[NAVIGATION]]", navigation)

    with open(htmlPath, "w") as f:
        f.write(html)



from __future__ import division

import logging

import pygame

from trosnoth.gui.common import TextImage

log = logging.getLogger(__name__)


class NameTag(pygame.sprite.Sprite):
    '''Sprite object that every player has which indicates the player's nick.'''
    def __init__(self, app, nick):
        pygame.sprite.Sprite.__init__(self)
        self.app = app

        if len(nick) > 15:
            nick = nick[:13] + '...'
        self.nick = nick
        colours = app.theme.colours
        nameFont = app.fonts.nameFont
        self.image = nameFont.render(app, self.nick, True,
                colours.nameTagShadow)
        foreground = nameFont.render(app, self.nick, True,
                colours.nameTagColour)
        self.image.blit(foreground, (-2, -2))

        self.image.set_colorkey((0,0,0))
        self.rect = self.image.get_rect()


class StarTally(pygame.sprite.Sprite):
    def __init__(self, app, stars):
        pygame.sprite.Sprite.__init__(self)
        self.app = app
        self.image = None
        self.rect = None
        self.stars = None

        self.setStars(stars)

    def setStars(self, stars):
        if self.stars == stars:
            return
        self.stars = stars

        pic = self.app.theme.sprites.smallStar
        starSize = pic.get_rect().size
        stars = max(stars, 0)

        flags = self.app.displaySettings.getSurfaceFlagsHack()

        if stars <= 5:
            self.image = pygame.Surface(
                (starSize[0] * stars, starSize[1]), flags)
            self.image.fill((255, 255, 255, 0))
            self.image.set_colorkey((255, 255, 255))

            # Blit the stars.
            for i in xrange(stars):
                self.image.blit(pic, (i * starSize[0], 0))

            self.rect = self.image.get_rect()
        else:
            self.image = pygame.Surface(
                (starSize[0] * 5, starSize[1] * 2), flags)
            self.image.fill((255, 255, 255, 0))
            self.image.set_colorkey((255, 255, 255))

            # Blit the stars.
            for i in xrange(5):
                self.image.blit(pic, (i * starSize[0] - 1, 0))
            for i in xrange(stars-5):
                self.image.blit(pic, (i * starSize[0] - 1, starSize[1]))

            self.rect = self.image.get_rect()


class HealthBar(pygame.sprite.Sprite):
    WIDTH = 30
    HEIGHT = 5

    def __init__(self, app, badColour, fairColour, goodColour):
        pygame.sprite.Sprite.__init__(self)
        self.app = app
        self.image = pygame.surface.Surface((self.WIDTH, self.HEIGHT)).convert()
        self.image.set_colorkey((0, 0, 1))
        self.rect = self.image.get_rect()

        self.badColour = badColour
        self.fairColour = fairColour
        self.goodColour = goodColour

        self.health = None
        self.maxHealth = None
        self.visible = False

    def setHealth(self, health, maxHealth):
        if (health, maxHealth) == (self.health, self.maxHealth):
            return
        self.health = health
        self.maxHealth = maxHealth
        if self.maxHealth == 1:
            self.visible = False
            return
        self.visible = True

        ratio = health / maxHealth
        if ratio <= 0.25 or health == 1:
            colour = self.badColour
        elif ratio <= 0.55:
            colour = self.fairColour
        else:
            colour = self.goodColour

        self.image.fill((0, 0, 1))
        r = pygame.Rect((0, 0), (self.WIDTH * ratio, self.HEIGHT))
        self.image.fill(colour, r)
        r.width = self.WIDTH
        pygame.draw.rect(
            self.image, self.app.theme.colours.healthBorder, r, 1)


class CountDown(pygame.sprite.Sprite):

    def __init__(
            self, app, player, font=None, colour=None, shadowColour=None,
            backColour=None):
        pygame.sprite.Sprite.__init__(self)

        if font is None:
            font = app.fonts.countdownFont
        if colour is None:
            colour = app.theme.colours.countDownColour
        if shadowColour is None:
            shadowColour = app.theme.colours.countDownShadow

        self.app = app
        self.player = player
        self.font = font
        self.colour = colour
        self.shadowColour = shadowColour

        self.foreText = TextImage(' ', font, colour, backColour, antialias=True)
        self.shadowText = TextImage('', font, shadowColour, backColour, antialias=True)

    def update(self):
        counter = int(
            5.99 * self.player.upgradeGauge / self.player.upgradeTotal)
        if counter == 0:
            text = ''
        else:
            text = str(counter)
        if text == self.foreText.text:
            return

        self.foreText.text = text
        self.shadowText.text = text

        image1 = self.foreText.getImage(self.app)
        image2 = self.shadowText.getImage(self.app)

        self.image = pygame.surface.Surface((
            image1.get_width() + 2, image1.get_height() + 2), pygame.SRCALPHA)
        self.rect = self.image.get_rect()

        self.image.blit(image2, (2, 2))
        self.image.blit(image1, (0, 0))



from trosnoth.gui.fonts import fontcache

class Font(object):
    def __init__(self, name, size, bold = False):
        self._name = name
        self._size = size
        self._bold = bold
        self._app = None
        self._filename = None

    def render(self, app, text, antialias, colour, background=None):
        if background==None:
            return self._getFont(app).render(text, antialias, colour)
        else:
            return self._getFont(app).render(text, antialias, colour,
                    background)

    def size(self, app, text):
        return self._getFont(app).size(text)

    def getLineSize(self, app):
        return self._getFont(app).get_linesize()

    def getHeight(self, app):
        return self._getFont(app).get_height()

    def _getFont(self, app):
        if app != self._app:
            filename = self._filename = app.getFontFilename(self._name)
            self._app = app
        else:
            filename = self._filename
        return fontcache.get(filename, self._size, self._bold)

    def __repr__(self):
        return "Font: %s size %d" % (self._name, self._size)

class ScaledFont(Font):
    def _getFont(self, app):
        if app != self._app:
            filename = self._filename = app.getFontFilename(self._name)
            self._app = app
        else:
            filename = self._filename
        return fontcache.get(filename,
                int(self._size * app.screenManager.scaleFactor + 0.5),
                self._bold)

    def __repr__(self):
        return "Scaled Font: %s size %d" % (self._name, self._size)



import logging
from math import sin, cos, atan2, sqrt

from trosnoth.const import SHOT_DEATH, GRENADE_DEATH
from trosnoth.model.unit import Unit, Bouncy, PredictedBouncyTrajectory
from trosnoth.utils.collision import collideTrajectory

log = logging.getLogger('shot')

GRENADE_BLAST_RADIUS = 448


class LocalSprite(object):
    '''
    Mixin for representing the local player's shots and grenades.
    These go slowly until their real counterpart catches up.
    '''

    def __init__(self, *args, **kwargs):
        super(LocalSprite, self).__init__(*args, **kwargs)
        self.nextPos = None
        self.localTicks = 0
        self.realTicks = 0
        self.realShotStarted = False
        self.realShotCaughtUp = False

    def advance(self):
        if self.realShotCaughtUp:
            super(LocalSprite, self).advance()
            return

        if self.realShotStarted:
            self.realTicks += 1

        if self.nextPos:
            self.pos = self.nextPos
            self.nextPos = None
            self.localTicks += 1
            if self.realTicks >= self.localTicks:
                self.realShotCaughtUp = True
        else:
            super(LocalSprite, self).advance()
            self.nextPos = self.pos
            self.pos = (
                0.5 * self.pos[0] + 0.5 * self.oldPos[0],
                0.5 * self.pos[1] + 0.5 * self.oldPos[1],
            )


class GrenadeShot(Bouncy):
    '''
    This will make the grenade have the same physics as a player without
    control and features of player movement
    '''

    HALF_WIDTH = 5
    HALF_HEIGHT = 5

    def __init__(self, world, player, duration, *args, **kwargs):
        Bouncy.__init__(self, world, *args, **kwargs)
        self.player = player
        self.timeLeft = duration

        # Place myself.
        self.pos = self.oldPos = player.pos
        angle = player.angleFacing
        self.xVel = self.world.physics.grenadeInitVel * sin(angle)
        self.yVel = -self.world.physics.grenadeInitVel * cos(angle)

    def getGravity(self):
        return self.world.physics.grenadeGravity

    def getMaxFallVel(self):
        return self.world.physics.grenadeMaxFallVel

    def advance(self):
        super(GrenadeShot, self).advance()
        self.timeLeft -= self.world.tickPeriod
        if self.timeLeft <= 0:
            xpos, ypos = self.pos
            self.propagateExplosionEvent()

            for player in self.world.players:
                if (
                        player.isFriendsWith(self.player) or player.phaseshift
                        or player.isInvulnerable() or player.turret
                        or player.dead):
                    continue
                dist = ((player.pos[0] - xpos) ** 2 +
                        (player.pos[1] - ypos) ** 2) ** 0.5
                if dist <= GRENADE_BLAST_RADIUS:
                    # Account for possibility player has left game
                    killer = (
                        self.player if self.player in self.world.players
                        else None)
                    player.zombieHit(killer, None, GRENADE_DEATH)
            self.removeGrenade()

    def propagateExplosionEvent(self):
        self.world.onGrenadeExplosion(self.pos, GRENADE_BLAST_RADIUS)

    def removeGrenade(self):
        self.world.removeGrenade(self)


class LocalGrenade(LocalSprite, GrenadeShot):
    def __init__(self, localState, *args, **kwargs):
        super(LocalGrenade, self).__init__(*args, **kwargs)
        self.localState = localState

    def propagateExplosionEvent(self):
        pass

    def removeGrenade(self):
        self.localState.grenadeRemoved()


class PredictedGrenadeTrajectory(PredictedBouncyTrajectory):

    HALF_WIDTH = GrenadeShot.HALF_WIDTH
    HALF_HEIGHT = GrenadeShot.HALF_HEIGHT

    def __init__(self, world, player, duration):
        PredictedBouncyTrajectory.__init__(self,
                                           world,
                                           player,
                                           duration,
                                           world.physics.grenadeInitVel,
                                           world.physics.grenadeGravity,
                                           world.physics.grenadeMaxFallVel)

    def explosionRadius(self):
        return GRENADE_BLAST_RADIUS


class Shot(Unit):

    NORMAL = 'normal'
    TURRET = 'turret'
    RICOCHET = 'ricochet'

    LIFETIME = 1.     # s

    HALF_WIDTH = 5
    HALF_HEIGHT = 5

    playerCollisionTolerance = 20

    def __init__(
            self, world, id, team, player, pos, velocity, kind, lifetime,
            *args, **kwargs):
        super(Shot, self).__init__(world, *args, **kwargs)

        self.id = id
        self.team = team
        self.pos = tuple(pos)
        self.originatingPlayer = player
        self.vel = tuple(velocity)
        self.timeLeft = lifetime
        self.kind = kind
        self.bounced = False
        self.expired = False
        self.justFired = True

    def isSolid(self):
        return self.kind != Shot.TURRET

    def checkCollision(self, player):
        if player.isFriendsWith(self.originatingPlayer):
            return False

        # Check both player colliding with us and us colliding with player
        if collideTrajectory(
                player.pos, self.oldPos,
                (self.pos[0] - self.oldPos[0], self.pos[1] - self.oldPos[1]),
                self.playerCollisionTolerance):
            return True

        deltaX = player.pos[0] - player.oldPos[0]
        deltaY = player.pos[1] - player.oldPos[1]
        if collideTrajectory(
                self.pos, player.oldPos, (deltaX, deltaY),
                self.playerCollisionTolerance):
            return True
        return False

    def collidedWithPlayer(self, player):
        self.expired = True
        player.onHitByShot(self)
        self.originatingPlayer.onShotHitSomething(self)
        if player.turret or player.isInvulnerable() or player.phaseshift:
            return
        player.zombieHit(self.originatingPlayer, self, SHOT_DEATH)
        self.originatingPlayer.onShotHurtPlayer(player, self)

    def advance(self):
        '''Called by the universe when this shot should update its position.
        deltaT is the time that's passed since its state was current, measured
        in seconds.'''
        if self.expired:
            return
        self.justFired = False

        deltaT = self.world.tickPeriod

        # Shots have a finite lifetime.
        self.timeLeft = self.timeLeft - deltaT
        if self.timeLeft <= 0:
            self.expired = True
            return

        dX, dY = self.vel[0] * deltaT, self.vel[1] * deltaT
        obstacle = self.world.physics.moveUnit(self, dX, dY)

        if obstacle is not None:
            # Shot hit an obstacle.
            if self.kind == Shot.RICOCHET:
                self.rebound(obstacle)
            else:
                self.expired = True
                return

    def rebound(self, obstacle):
        '''
        Shot is a ricochet shot and it's hit an obstacle.
        '''
        self.bounced = True
        obsAngle = obstacle.getAngle()
        shotAngle = atan2(self.vel[1], self.vel[0])
        dif = shotAngle - obsAngle
        final = obsAngle - dif
        speed = sqrt(self.vel[0] * self.vel[0] + self.vel[1] * self.vel[1])
        self.vel = (speed * cos(final), speed * sin(final))


class LocalShot(LocalSprite, Shot):
    def advance(self):
        self.justFired = False
        super(LocalShot, self).advance()



from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns('',
    url(r'^$', views.index, name='index'),
    url(r'^u/?$', views.userList, name='userlist'),
    url(r'^u/(?P<userId>\d+)/?$', views.userProfile, name='profile'),
    url(
        r'^u/(?P<userId>\d+)/(?P<nick>[^/]+)/?$',
        views.userProfile, name='profile'),
    url(r'^g/?$', views.gameList, name='gamelist'),
    url(r'^g/(?P<gameId>\d+)/?$', views.viewGame, name='viewgame'),
)




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2009  Joshua Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

'''statusBar.py - defines the StatusBar class which deals with drawing the
zone tallies onto the screen.'''

import random
import logging
import pygame

from trosnoth.gui.framework.elements import TextElement
from trosnoth.gui.common import Location
import trosnoth.gui.framework.framework as framework
from trosnoth.gui.framework.basics import Animation
from trosnoth.utils.utils import timeNow

log = logging.getLogger(__name__)

class GameProgressBar(framework.CompoundElement):
    def __init__(self, app, world, gameViewer):
        super(GameProgressBar, self).__init__(app)
        self.world = world
        self.gameViewer = gameViewer
        self.app = app
        self.disrupt = False
        self.disruptTick = 0

        colours = app.theme.colours
        self.black = colours.black
        self.blue = colours.team1Mn_zone
        self.red = colours.team2Mn_zone
        self.grey = colours.zoneBarNeutral

        self.barFont = app.screenManager.fonts.zoneBarFont

        # Define a few constants to make things easier
        self.triangleLength = 25
        self.sideDistance = 4
        self.barHeight = self.gameViewer.zoneBarHeight
        self.textSpace = 130
        self.neutralTextSpace = 15

        self._oldRect = None
        self._oldScale = None
        self.width = None
        self.mapLeft = None
        self.mapRight = None
        self.mapBottom = None
        self.xFarLeft = None
        self.xFarRight = None
        self.xLeft = None
        self.xRight = None
        self.yTop = None
        self.yBottom = None
        self.disruptAnimation = None
        self.calculateSize()

    def calculateSize(self):
        minimapRect = self.gameViewer.miniMap.getRect()
        scaleFactor = self.app.screenManager.scaleFactor
        if minimapRect == self._oldRect and scaleFactor == self._oldScale:
            return
        self._oldRect = minimapRect
        self._oldScale = scaleFactor

        # Width will always be between 1 and 2 inclusive
        self.width = min(max(int(3 * scaleFactor), 1), 2)

        self.mapLeft = minimapRect.left
        self.mapRight = minimapRect.right
        self.mapBottom = minimapRect.bottom
        self.xFarLeft = self.mapLeft + self.sideDistance
        self.xFarRight = self.mapRight - self.sideDistance

        self.xLeft = self.xFarLeft + self.triangleLength
        self.xRight = self.xFarRight - self.triangleLength

        self.yTop = self.mapBottom - 1
        self.yBottom = self.yTop + self.barHeight

        # Create the disrupted animation
        self.disruptAnimation = self.createDisrupted()

    def draw(self, surface):
        self.calculateSize()
        super(GameProgressBar, self).draw(surface)

        if self.disrupt and random.random() < 0.2:
            disrupt = True
        else:
            disrupt = False

        # Automatically generated constants
        yTop = self.yTop
        yBottom = self.yBottom
        yText = yTop

        xFarLeft = self.xFarLeft
        xFarRight = self.xFarRight

        xLeft = self.xLeft
        xRight = self.xRight

        # Define the coordinates for the static shapes
        border = [
            (xFarLeft, yTop),
            (xFarRight, yTop),
            (xRight, yBottom),
            (xLeft, yBottom),
        ]

        blueTriangle = [
            (xFarLeft, yTop),
            (xLeft, yTop),
            (xLeft, yBottom),
        ]

        redTriangle = [
            (xFarRight, yTop),
            (xRight, yTop),
            (xRight, yBottom),
        ]

        # Get the information we need
        blueScore = self.getBlueScore()
        blueProportion = self.getBlueProportion()
        redScore = self.getRedScore()
        redProportion = self.getRedProportion()
        neutralScore = self.getNeutralScore()
        neutralProportion = self.getNeutralProportion()

        gameOver = self.isGameOver()

        # Define the coordinates for the dynamic shapes
        mutableBarLength = xRight - xLeft
        if redProportion > 0:
            blueBarLength = int(round(mutableBarLength * blueProportion))
        else:
            blueBarLength = mutableBarLength
        if blueProportion > 0:
            redBarLength = int(round(mutableBarLength * redProportion))
        else:
            redBarLength = mutableBarLength

        xNeutral = ((xLeft + blueBarLength) + (xRight - redBarLength)) / 2

        xFirstBar = xLeft + blueBarLength
        xSecondBar = xRight - redBarLength

        blueBar = [
            (xLeft, yTop),
            (xLeft + blueBarLength, yTop),
            (xLeft + blueBarLength, yBottom),
            (xLeft, yBottom),
        ]
        redBar = [
            (xRight - redBarLength, yTop),
            (xRight, yTop),
            (xRight, yBottom),
            (xRight - redBarLength, yBottom),
        ]
        greyBar = [
            (xLeft + blueBarLength + 1, yTop),
            (xRight - redBarLength - 1, yTop),
            (xRight - redBarLength - 1, yBottom),
            (xLeft + blueBarLength + 1, yBottom),
        ]

        # Draw the two triangles on the sides
        if blueProportion > 0:
            pygame.draw.polygon(surface, self.blue, blueTriangle, 0)
        elif redProportion > 0:
            pygame.draw.polygon(surface, self.red, blueTriangle, 0)
        else:
            pygame.draw.polygon(surface, self.grey, blueTriangle, 0)

        if redProportion > 0:
            pygame.draw.polygon(surface, self.red, redTriangle, 0)
        elif blueProportion > 0:
            pygame.draw.polygon(surface, self.blue, redTriangle, 0)
        else:
            pygame.draw.polygon(surface, self.grey, redTriangle, 0)

        # Draw the team colours
        if blueProportion > 0:
            pygame.draw.polygon(surface, self.blue, blueBar, 0)
        if redProportion > 0:
            pygame.draw.polygon(surface, self.red, redBar, 0)
        if neutralProportion > 0 and (blueScore > 0) == (redScore > 0):
            allNeutral = True
            pygame.draw.polygon(surface, self.grey, greyBar, 0)
        else:
            allNeutral = False

        # Draw the black seperator line(s)
        if not gameOver:
            pygame.draw.line(surface, self.black, (xFirstBar, yTop),
                    (xFirstBar, yBottom), self.width)
            if neutralScore != 0:
                pygame.draw.line(surface, self.black, (xSecondBar, yTop),
                        (xSecondBar, yBottom), self.width)

        # Draw the disruption
        if disrupt:
            surface.blit(self.disruptAnimation.getImage(), (self.xFarLeft,
                    self.yTop))

        # Draw the border last so that it goes on top
        colours = self.app.theme.colours
        pygame.draw.polygon(surface, self.black, border, self.width)
        pygame.draw.line(surface, colours.minimapBorder, (self.mapLeft,
                self.mapBottom - 1), (self.mapRight, self.mapBottom - 1), 2)

        # Define the necessary text
        self.blueText = TextElement(self.app, '', self.barFont,
                Location((xFirstBar - 5, yText), 'topright'),
                colour=colours.black)
        self.neutralText = TextElement(self.app, '', self.barFont,
                Location((xNeutral, yText), 'midtop'), colour=colours.black)
        self.redText = TextElement(self.app, '', self.barFont,
                Location((xSecondBar + 7, yText), 'topleft'),
                colour=colours.black)

        if not disrupt:
            blueString = self.getScoreText(blueScore)
            redString = self.getScoreText(redScore)

            if xSecondBar - xFirstBar > self.neutralTextSpace or allNeutral:
                neutralString = self.getScoreText(neutralScore)
            else:
                neutralString = ""

            self.blueText.setText(blueString)
            self.redText.setText(redString)
            self.neutralText.setText(neutralString)
        else:
            self.blueText.setText('')
            self.redText.setText('')
            self.neutralText.setText('')

        # Draw the text
        self.blueText.draw(surface)
        self.redText.draw(surface)
        self.neutralText.draw(surface)

    def createDisrupted(self):
        screens = []
        for a in range(0,4):
            screen = (pygame.surface.Surface((self.xFarRight - self.xFarLeft,
                                              self.gameViewer.zoneBarHeight)))
            x = y = 0
            xL = 1
            xR = self.xFarRight - self.xFarLeft - 1
            rect = pygame.rect.Rect(0,0,2,2)
            disruptColours = [self.red, self.blue]
            while y < self.gameViewer.zoneBarHeight:
                while x < xR:
                    rect.left = x
                    rect.top = y
                    pygame.draw.rect(screen, random.choice(disruptColours),
                            rect, 0)
                    x += 2
                xL += 2
                xR -= 2
                x = xL
                y += 2
            screen.set_colorkey((0,0,0))
            screens.append(screen)
        return Animation(0.1, timeNow, *screens)

    def getScoreText(self, score):
        raise NotImplementedError

    def getBlueScore(self):
        raise NotImplementedError

    def getRedScore(self):
        raise NotImplementedError

    def getNeutralScore(self):
        raise NotImplementedError

    def getBlueProportion(self):
        raise NotImplementedError

    def getNeutralProportion(self):
        raise NotImplementedError

    def getRedProportion(self):
        raise NotImplementedError

    def isGameOver(self):
        raise NotImplementedError

class ZoneProgressBar(GameProgressBar):
    def getScoreText(self, score):
        if score == 0:
            return ''
        return str(score)

    def getBlueScore(self):
        return self.world.teams[0].numZonesOwned

    def getRedScore(self):
        return self.world.teams[1].numZonesOwned

    def getNeutralScore(self):
        return len(self.world.zones) - self.getBlueScore() - self.getRedScore()

    def getBlueProportion(self):
        return self.getBlueScore() / (len(self.world.zones) + 0.0)

    def getNeutralProportion(self):
        #if self.isGameOver():
        #    return 0.0
        return self.getNeutralScore() / (len(self.world.zones) + 0.0)

    def getRedProportion(self):
        return self.getRedScore() / (len(self.world.zones) + 0.0)

    def isGameOver(self):
        return self.getBlueScore() == 0 or self.getRedScore() == 0

class FrontLineProgressBar(GameProgressBar):
    def getScoreText(self, score):
        return str(score)

    def getBlueScore(self):
        return str(self.world.trosballScore[self.world.teams[0].id])

    def getRedScore(self):
        return str(self.world.trosballScore[self.world.teams[1].id])

    def getNeutralScore(self):
        return 0

    def getBlueProportion(self):
        frontLine = self.world.uiOptions.frontLineGetter()
        mapWidth = self.world.layout.worldSize[0]
        return max(0.0, min(1.0, frontLine / (mapWidth + 0.0)))

    def getNeutralProportion(self):
        return 0.0

    def getRedProportion(self):
        return 1 - self.getBlueProportion()

    def isGameOver(self):
        return False



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging
import os

from trosnoth.const import (
    OPEN_CHAT, GAME_FULL_REASON, UNAUTHORISED_REASON, NICK_USED_REASON,
    USER_IN_GAME_REASON, ALREADY_JOINED_REASON,
)
from trosnoth.model.agent import ConcreteAgent
from trosnoth.messages import (
    TaggingZoneMsg, ShootMsg, RespawnMsg, RespawnRequestMsg, TickMsg,
    CannotJoinMsg, UpdatePlayerStateMsg,
    AimPlayerAtMsg, BuyUpgradeMsg, ChatMsg,
)
from trosnoth.model.universe_base import NeutralTeamId
from trosnoth.utils.message import MessageConsumer
from trosnoth.utils.twist import WeakCallLater, WeakLoopingCall
import trosnoth.ais
from math import atan2

log = logging.getLogger('ai')


def listAIs(playableOnly=False):
    '''
    Returns a list of strings of the available AI classes.
    '''
    results = []
    try:
        files = os.listdir(os.path.dirname(trosnoth.ais.__file__))
    except OSError:
        # Probably frozen in a zip. Use a standard list.
        aiNames = ['alpha', 'simple', 'john', 'test']
    else:
        aiNames = [os.path.splitext(f)[0] for f in files if f.endswith('.py')]

    for aiName in aiNames:
        c = __import__('trosnoth.ais.%s' % (aiName,), fromlist=['AIClass'])
        if hasattr(c, 'AIClass') and (
                c.AIClass.playable or not playableOnly):
            results.append(aiName)
    return results


def makeAIAgent(game, aiName):
    c = __import__('trosnoth.ais.%s' % (aiName,), fromlist=['AIClass'])
    return AIAgent(game, c.AIClass)


class AIAgent(ConcreteAgent):
    '''
    Base class for an AI agent.
    '''

    def __init__(self, game, aiClass, *args, **kwargs):
        super(AIAgent, self).__init__(game=game, *args, **kwargs)
        self.aiClass = aiClass
        self.ai = None
        self.team = None
        self.requestedNick = None
        self._loop = WeakLoopingCall(self, '_tick')

    def start(self, team=None):
        self.team = team
        self._loop.start(2)

    def stop(self):
        super(AIAgent, self).stop()
        self._loop.stop()

    def _tick(self):
        if self.ai is not None:
            return
        if not self.world.state.aisEnabled():
            return

        nick = self.aiClass.nick

        if self.team is None:
            teamId = NeutralTeamId
        else:
            teamId = self.team.id

        self._joinGame(nick, teamId)

    def _joinGame(self, nick, teamId):
        self.requestedNick = nick
        nick = '%s-1' % (nick,)
        self.sendJoinRequest(teamId, nick, bot=True)

    @CannotJoinMsg.handler
    def _joinFailed(self, msg):
        r = msg.reasonId
        nick = self.requestedNick

        if r == GAME_FULL_REASON:
            message = 'full'
        elif r == UNAUTHORISED_REASON:
            message = 'not authenticated'
        elif r == NICK_USED_REASON:
            message = 'nick in use'
        elif r == USER_IN_GAME_REASON:
            message = 'user already in game'    # Should never happen
        elif r == ALREADY_JOINED_REASON:
            message = 'tried to join twice'     # Should never happen
        else:
            message = repr(r)

        log.error('Join failed for AI %r (%s)', nick, message)
        self.stop()

    def setPlayer(self, player):
        if player is None:
            log.debug('DISABLE AI: %s', self.ai)
            self.ai.disable()
            self.ai = None

        super(AIAgent, self).setPlayer(player)

        if player:
            self.requestedNick = None
            self.ai = self.aiClass(self.world, self.localState.player, self)

    @TickMsg.handler
    def handle_TickMsg(self, msg):
        super(AIAgent, self).handle_TickMsg(msg)
        if self.ai:
            self.ai.consumeMsg(msg)

    def defaultHandler(self, msg):
        super(AIAgent, self).defaultHandler(msg)
        if self.ai:
            self.ai.consumeMsg(msg)


class AI(MessageConsumer):
    playable = False    # Change to True in playable subclasses.

    def __init__(self, world, player, agent, *args, **kwargs):
        super(AI, self).__init__(*args, **kwargs)
        self.world = world
        self.player = player
        self.player.onDied.addListener(self.died)
        self.agent = agent
        self.start()

    def sendRequest(self, msg):
        self.agent.sendRequest(msg)

    def start(self):
        raise NotImplementedError

    def disable(self):
        raise NotImplementedError

    def defaultHandler(self, msg):
        pass

    def _sendStateUpdate(self, key, value):
        self.sendRequest(UpdatePlayerStateMsg(
            value, stateKey=key, tickId=self.world.lastTickId))

    def doMoveRight(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('right', True)
        self._sendStateUpdate('left', False)

    def doMoveLeft(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('left', True)
        self._sendStateUpdate('right', False)

    def doStop(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('left', False)
        self._sendStateUpdate('right', False)

    def doDrop(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('down', True)
        WeakCallLater(0.1, self, '_releaseDropKey')

    def _releaseDropKey(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('down', False)

    def doJump(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('jump', True)

    def doStopJump(self):
        if self.alreadyRemoved():
            return
        self._sendStateUpdate('jump', False)

    def doAimAt(self, angle, thrust=1.0):
        if self.alreadyRemoved():
            return
        self.sendRequest(AimPlayerAtMsg(angle, thrust, self.world.lastTickId))

    def doAimAtPoint(self, pos, thrust=1.0):
        if self.alreadyRemoved():
            return
        x1, y1 = self.player.pos
        x2, y2 = pos
        angle = atan2(x2 - x1, -(y2 - y1))
        self.doAimAt(angle, thrust)

    def doShoot(self):
        if self.alreadyRemoved():
            return
        self.sendRequest(ShootMsg(self.world.lastTickId))

    def doRespawn(self):
        '''
        Attempts to respawn.
        '''
        if self.alreadyRemoved():
            return
        self.sendRequest(RespawnRequestMsg(self.world.lastTickId))

    def doBuyUpgrade(self, upgradeKind):
        '''
        Attempts to purchase an upgrade of the given kind.
        '''
        if self.alreadyRemoved():
            return
        self.sendRequest(BuyUpgradeMsg(
            upgradeKind.upgradeType, self.world.lastTickId))

    @RespawnMsg.handler
    def _respawned(self, msg):
        if msg.playerId == self.player.id:
            self.respawned()

    def respawned(self):
        pass

    def died(self, killer, deathType):
        pass

    @TaggingZoneMsg.handler
    def _zoneTagged(self, msg):
        self.zoneTagged(msg.zoneId, msg.playerId)

    def zoneTagged(self, zoneId, playerId):
        pass

    def doSendPublicChat(self, message):
        if self.alreadyRemoved():
            return
        self.sendRequest(ChatMsg(OPEN_CHAT, text=message.encode()))

    def alreadyRemoved(self):
        if self.player.id == -1:
            log.warning(
                'AI trying to continue after being removed: %s' %
                (self.__class__.__name__,))
            return True
        return False



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


from trosnoth.gui.framework.menu import MenuDisplay
from trosnoth.gui.menu.menu import MenuManager, Menu, MenuItem
from trosnoth.model.upgrades import allUpgrades
from trosnoth.gui.common import Size

class MainMenu(MenuDisplay):
    def __init__(self, app, location, interface, keymapping):
        font = app.screenManager.fonts.ingameMenuFont
        titleColour = (255, 255, 255)
        stdColour = (255, 255, 0)
        hvrColour = (0, 255, 255)
        backColour = (0, 64, 192)
        autosize = True
        hidable = True
        size = Size(175, 10)   # Height doesn't matter when autosize is set.

        self.ACCELERATION = 1000    # pix/s/s

        manager = MenuManager()
        upgrades = [MenuItem('%s (%s)' % (upgradeClass.name,
                upgradeClass.requiredStars), upgradeClass.action) for
                upgradeClass in sorted(allUpgrades,
                key=lambda upgradeClass: upgradeClass.order)]
        self.buyMenu = Menu(name='Select Upgrade', listener=interface.doAction,
                items=upgrades + [
            MenuItem('Deselect upgrade', 'no upgrade'),
            MenuItem('Cancel', 'menu')
        ])

        self.moreMenu = Menu(name='More Actions', listener=interface.doAction,
                items=[
            MenuItem('Abandon upgrade', 'abandon'),
            MenuItem('Chat', 'chat'),
            MenuItem('Show/hide HUD', 'toggle interface'),
            MenuItem('Cancel', 'menu')
        ])
        self.quitMenu = Menu(name='Really Quit?', listener=interface.doAction,
                items=[
            MenuItem('Leave game', 'leave'),
            MenuItem('---'),
            MenuItem('Cancel', 'menu')
        ])
        self.playMenu = Menu(name='Menu', listener=interface.doAction, items=[
            MenuItem('Respawn', 'respawn'),
            MenuItem('Select upgrade...', 'select upgrade'),
            MenuItem('Activate upgrade', 'activate upgrade'),
            MenuItem('Change nickname', 'change nickname'),
            MenuItem('Settings', 'settings'),
            MenuItem('More...', 'more actions'),
            MenuItem('---'),
            MenuItem('Leave game', 'maybeLeave')
        ])
        self.replayMenu = Menu(
            name='Menu', listener=interface.doAction,
            items=[
                MenuItem('Settings', 'settings'),
                MenuItem('---'),
                MenuItem('Quit', 'maybeLeave')
            ],
        )
        self.spectateMenu = Menu(
            name='Menu', listener=interface.doAction,
            items=[
                MenuItem('Join...', 'join'),
                MenuItem('Settings', 'settings'),
                MenuItem('---'),
                MenuItem('Quit', 'maybeLeave')
            ],
        )

        manager.setDefaultMenu(self.playMenu)

        super(MainMenu, self).__init__(app, location, size, font, manager,
                titleColour, stdColour, hvrColour, None, backColour, autosize,
                hidable, keymapping)

    def setMode(self, replay=False, spectate=False):
        if replay:
            self.manager.setDefaultMenu(self.replayMenu)
        elif spectate:
            self.manager.setDefaultMenu(self.spectateMenu)
        else:
            self.manager.setDefaultMenu(self.playMenu)
        self.manager.reset()

    def showBuyMenu(self):
        self.manager.reset()
        self.manager.showMenu(self.buyMenu)

    def showMoreMenu(self):
        self.manager.reset()
        self.manager.showMenu(self.moreMenu)

    def showQuitMenu(self):
        self.manager.reset()
        self.manager.showMenu(self.quitMenu)

    def escape(self):
        if self.hidden:
            # Just show the existing menu.
            self.hide()
        elif self.manager.menu == self.manager.defaultMenu:
            # Main menu is already selected. Hide it.
            self.hide()
        else:
            # Main menu is not selected. Return to it.
            self.manager.cancel()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame
from twisted.internet import defer

from trosnoth.gui.framework.dialogbox import DialogBox
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.framework.tabContainer import TabContainer
from trosnoth.gui.common import Region, Canvas, Location
from trosnoth.gui.framework.elements import TextElement, TextButton
from trosnoth.gui.framework import prompt, framework
from trosnoth.utils.twist import WeakCallLater

class PasswordGUIError(Exception):
    pass

class PasswordGUI(DialogBox):
    def __init__(self, app):
        size = Canvas(512, 384)
        DialogBox.__init__(self, app, size, 'Please authenticate')
        self._deferred = None
        self._host = None

        font = app.screenManager.fonts.defaultTextBoxFont
        btnColour = app.theme.colours.dialogButtonColour
        highlightColour = app.theme.colours.black
        errorTextColour = app.theme.colours.errorColour

        self.tabContainer = TabContainer(app,
                Region(topleft=self.Relative(0, 0),
                size=self.Relative(1, 0.75)), font,
                app.theme.colours.tabContainerColour)
        self.tabContainer.addTab(LoginTab(app))
        self.tabContainer.addTab(CreateAccountTab(app))

        self.errorText = TextElement(app, '', font,
            Location(self.Relative(0.5, 0.8), 'center'), errorTextColour)

        font = app.screenManager.fonts.bigMenuFont
        self.elements = [
            self.tabContainer,
            self.errorText,
            TextButton(app,
                Location(self.Relative(0.3, 0.9), 'center'),
                'Ok', font, btnColour, highlightColour,
                onClick=self.okClicked),
            TextButton(app,
                Location(self.Relative(0.7, 0.9), 'center'),
                'Cancel', font, btnColour, highlightColour,
                onClick=self.cancelClicked),
        ]
        self.cancelled = False

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN and event.key in (pygame.K_KP_ENTER,
                pygame.K_RETURN):
            self.okClicked()
            return None
        else:
            return DialogBox.processEvent(self, event)

    def cancelClicked(self, element=None):
        self.close()
        self._deferred.callback(None)

    def okClicked(self, element=None):
        if self.tabContainer.selectedIndex == 1:
            create = True
            # Check that passwords match.
            tab = self.tabContainer.tabs[1]
            if tab.passwordField.value != tab.passwordField2.value:
                self.setErrorText('Passwords do not match!')
                return
        else:
            create = False
            tab = self.tabContainer.tabs[0]

        username = tab.usernameField.value
        password = tab.passwordField.value

        if len(username) == 0:
            self.setErrorText('You must give a username!')
            return
        if len(password) == 0:
            self.setErrorText('Password cannot be blank!')
            return

        self.close()
        self.app.identitySettings.usernames[self._host] = username
        self._deferred.callback((create, username, password))

    def getPassword(self, host, errorMsg=''):
        if self.showing:
            raise PasswordGUIError('PasswordGUI already showing')
        if self.cancelled:
            self.cancelled = False
            result = self._deferred = defer.Deferred()
            WeakCallLater(0.1, result, 'callback', None)
            return result
        self.setCaption(host)
        self.tabContainer.tabs[0].reset(host)
        self.tabContainer.tabs[1].reset()
        self.setErrorText(errorMsg)
        self.show()
        self._host = host
        result = self._deferred = defer.Deferred()
        return result

    def setErrorText(self, text):
        self.errorText.setText(text)

class LoginTab(Tab, framework.TabFriendlyCompoundElement):
    def __init__(self, app):
        Tab.__init__(self, app, 'Sign in')
        framework.TabFriendlyCompoundElement.__init__(self, app)
        font = app.screenManager.fonts.defaultTextBoxFont
        labelColour = app.theme.colours.dialogBoxTextColour
        inputColour = app.theme.colours.grey

        self.usernameField = prompt.InputBox(app,
            Region(topleft=self.Relative(0.1, 0.16),
                bottomright=self.Relative(0.9, 0.32)),
                font=font, colour=inputColour,
                validator=usernameValidator,
                onClick=self.setFocus, onTab=self.tabNext)

        self.passwordField = prompt.PasswordBox(app,
            Region(topleft=self.Relative(0.1, 0.48),
                bottomright=self.Relative(0.9, 0.64)),
                font=font, colour=inputColour,
                onClick=self.setFocus, onTab=self.tabNext)

        self.elements = [
            TextElement(app, 'Username', font,
                Location(self.Relative(0.1, 0.11), 'midleft'),
                labelColour),
            self.usernameField,
            TextElement(app, 'Password', font,
                Location(self.Relative(0.1, 0.43), 'midleft'),
                labelColour),
            self.passwordField,
        ]
        self.tabOrder = [self.usernameField, self.passwordField]

    def reset(self, host):
        self.passwordField.setValue('')
        self.setFocus(self.passwordField)

        if self.usernameField.value == '':
            username = self.app.identitySettings.usernames.get(host)
            if username is None:
                self.setFocus(self.usernameField)
            else:
                self.usernameField.setValue(username)

class CreateAccountTab(Tab, framework.TabFriendlyCompoundElement):
    def __init__(self, app):
        Tab.__init__(self, app, 'New account')
        framework.TabFriendlyCompoundElement.__init__(self, app)
        font = app.screenManager.fonts.defaultTextBoxFont
        labelColour = app.theme.colours.dialogBoxTextColour
        inputColour = app.theme.colours.grey

        self.usernameField = prompt.InputBox(app,
            Region(topleft=self.Relative(0.1, 0.16),
                bottomright=self.Relative(0.9, 0.32)),
                font=font, colour=inputColour,
                validator=usernameValidator,
                onClick=self.setFocus, onTab=self.tabNext)

        self.passwordField = prompt.PasswordBox(app,
            Region(topleft=self.Relative(0.1, 0.48),
                bottomright=self.Relative(0.9, 0.64)),
                font=font, colour=inputColour,
                onClick=self.setFocus, onTab=self.tabNext)

        self.passwordField2 = prompt.PasswordBox(app,
            Region(topleft=self.Relative(0.1, 0.8),
                bottomright=self.Relative(0.9, 0.96)),
                font=font, colour=inputColour,
                onClick=self.setFocus, onTab=self.tabNext)

        self.elements = [
            TextElement(app, 'Username', font,
                Location(self.Relative(0.1, 0.11), 'midleft'),
                labelColour),
            self.usernameField,
            TextElement(app, 'Password', font,
                Location(self.Relative(0.1, 0.43), 'midleft'),
                labelColour),
            self.passwordField,
            TextElement(app, 'Retype password', font,
                Location(self.Relative(0.1, 0.75), 'midleft'),
                labelColour),
            self.passwordField2,
        ]
        self.tabOrder = [self.usernameField, self.passwordField,
                self.passwordField2]

    def reset(self):
        self.passwordField.setValue('')
        self.passwordField2.setValue('')
        self.setFocus(self.usernameField)


USERNAME_CHARS = set(
        'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789-')
def usernameValidator(char):
    return char in USERNAME_CHARS



import pygame

from trosnoth.gui.framework import framework, prompt
from trosnoth.gui.framework.elements import TextElement
from trosnoth.gui.framework.checkbox import CheckBox
from trosnoth.gui.framework.slider import Slider
from trosnoth.gui.framework.tab import Tab
from trosnoth.trosnothgui.common import button
from trosnoth.gui.common import ScaledLocation, ScaledArea
from trosnoth.utils.event import Event


class DisplaySettingsTab(Tab, framework.TabFriendlyCompoundElement):
    def __init__(self, app, onClose=None):
        super(DisplaySettingsTab, self).__init__(app, 'Display')

        self.onClose = Event()
        if onClose is not None:
            self.onClose.addListener(onClose)

        font = self.app.screenManager.fonts.bigMenuFont
        smallNoteFont = self.app.screenManager.fonts.smallNoteFont

        colour = self.app.theme.colours.headingColour
        def mkText(text, x, y, textFont=font, anchor='topright'):
            return TextElement(self.app, text, textFont,
                    ScaledLocation(x, y, anchor),
                    colour)

        self.text = [
            mkText('X', 640, 280),
            mkText('Screen resolution', 430, 280),
            mkText('Fullscreen mode', 430, 360),
            mkText('Graphics detail', 430, 440),
            mkText('low', 460, 475, textFont=smallNoteFont, anchor='midtop'),
            mkText('high', 845, 475, textFont=smallNoteFont, anchor='midtop'),
            mkText('Show timings', 430, 525),
        ]

        self.invalidInputText = TextElement(self.app, '', font,
                ScaledLocation(512, 230,'midtop'), (192, 0, 0))

        self.widthInput = prompt.InputBox(self.app,
                ScaledArea(460, 265, 150, 60),
                initValue=str(self.app.screenManager.size[0]), font=font,
                maxLength=4, validator=prompt.intValidator)

        self.widthInput.onEnter.addListener(lambda sender: self.saveSettings())
        self.widthInput.onClick.addListener(self.setFocus)
        self.widthInput.onTab.addListener(self.tabNext)

        self.heightInput = prompt.InputBox(self.app,
                ScaledArea(652, 265, 150, 60),
                initValue=str(self.app.screenManager.size[1]), font=font,
                maxLength=4, validator=prompt.intValidator)

        self.heightInput.onEnter.addListener(lambda sender: self.saveSettings())
        self.heightInput.onClick.addListener(self.setFocus)
        self.heightInput.onTab.addListener(self.tabNext)

        self.tabOrder = [self.widthInput, self.heightInput]

        self.fullscreenBox = CheckBox(self.app,
            ScaledLocation(460, 365),
            text='',
            font=font,
            colour=(192, 192, 192),
            initValue=self.app.screenManager.isFullScreen(),
        )
        self.fullscreenBox.onValueChanged.addListener(self.fullscreenChanged)

        displaySettings = app.displaySettings

        self.detailSlider = Slider(
            self.app, ScaledArea(460, 430, 390, 40),
            bounds=(0, len(displaySettings.DETAIL_LEVELS) - 1), snap=True)
        self.detailSlider.setVal(
            displaySettings.DETAIL_LEVELS.index(displaySettings.detailLevel))
        self.detailSlider.onValueChanged.addListener(self.detailChanged)

        self.showTimingsBox = CheckBox(self.app,
            ScaledLocation(460, 530),
            text='',
            font=font,
            colour=(192,192,192),
            initValue=displaySettings.showTimings,
        )

        self.input = [self.widthInput, self.heightInput, self.widthInput,
                self.fullscreenBox, self.detailSlider, self.showTimingsBox]

        self.elements = self.text + self.input + [
            self.invalidInputText,
            button(app, 'save', self.saveSettings, (-100, -75), 'midbottom',
                    secondColour=app.theme.colours.white),
            button(app, 'cancel', self.cancelMenu, (100, -75), 'midbottom',
                    secondColour=app.theme.colours.white),
        ]
        self.setFocus(self.widthInput)

    def detailChanged(self, newLevel):
        pass

    def cancelMenu(self):
        self.fullscreenBox.setValue(self.app.screenManager.isFullScreen())
        self.showTimingsBox.setValue(self.app.displaySettings.showTimings)
        self.heightInput.setValue(str(self.app.screenManager.size[1]))
        self.widthInput.setValue(str(self.app.screenManager.size[0]))

        self.onClose.execute()

    def saveSettings(self):
        displaySettings = self.app.displaySettings

        height = self.getInt(self.heightInput.value)
        width = self.getInt(self.widthInput.value)
        fullScreen = self.fullscreenBox.value
        detailLevel = displaySettings.DETAIL_LEVELS[self.detailSlider.getVal()]
        showTimings = self.showTimingsBox.value

        # The resolutionList is used when fullScreen is true.
        resolutionList = pygame.display.list_modes()
        resolutionList.sort()
        minResolution = resolutionList[0]
        maxResolution = resolutionList[-1]

        if not fullScreen:
            minResolution = (320, 240)

        # These values are used when fullScreen is false.
        widthRange = (minResolution[0], maxResolution[0])
        heightRange = (minResolution[1], maxResolution[1])

        if not widthRange[0] <= width <= widthRange[1]:
            self.incorrectInput('Screen width must be between %d and %d' %
                                (widthRange[0], widthRange[1]))
            width = None
            return
        if not heightRange[0] <= height <= heightRange[1]:
            self.incorrectInput('Screen height must be between %d and %d' %
                                (heightRange[0], heightRange[1]))
            height = None
            return
        if fullScreen:
            selectedResolution = (width, height)
            if selectedResolution not in resolutionList:
                self.incorrectInput('Selected resolution is not valid for '
                        'this display')
                height = width = None
                return

        self.incorrectInput('')

        # Save these values.
        displaySettings.fullScreen = fullScreen
        displaySettings.detailLevel = detailLevel
        displaySettings.showTimings = showTimings
        if fullScreen:
            displaySettings.fsSize = (width, height)
        else:
            displaySettings.size = (width, height)

        # Write to file and apply.
        displaySettings.save()
        displaySettings.apply()

        self.onClose.execute()

    def getInt(self, value):
        if value == '':
            return 0
        return int(value)

    def incorrectInput(self, string):
        self.invalidInputText.setText(string)
        self.invalidInputText.setFont(self.app.screenManager.fonts.bigMenuFont)

    def fullscreenChanged(self, element):
        # If the resolution boxes haven't been touched, swap their values to
        # the appropriate resolution for the new mode.

        height = self.getInt(self.heightInput.value)
        width = self.getInt(self.widthInput.value)
        fullScreen = self.fullscreenBox.value

        if fullScreen:
            # Going to full screen mode.
            if (width, height) != self.app.displaySettings.size:
                return
            width, height = self.app.displaySettings.fsSize
        else:
            # Going from full screen mode.
            if (width, height) != self.app.displaySettings.fsSize:
                return
            width, height = self.app.displaySettings.size

        self.heightInput.setValue(str(height))
        self.widthInput.setValue(str(width))



'''viewManager.py - defines the ViewManager class which deals with drawing the
state of a universe to the screen.'''

import math
import logging
import random

import pygame

from trosnoth.const import (
    MAP_TO_SCREEN_SCALE, BODY_BLOCK_SCREEN_SIZE, INTERFACE_BLOCK_SCREEN_SIZE,
)
from trosnoth.model.player import Player
from trosnoth.utils import globaldebug
from trosnoth.utils.utils import timeNow
from trosnoth.themes import BLOCK_BACKGROUND_COLOURKEY
from trosnoth.trosnothgui.ingame.minimap import MiniMap
from trosnoth.trosnothgui.ingame.utils import (
    mapPosToScreen, screenToMapPos, viewRectToMap,
)
from trosnoth.gui.framework import framework
from trosnoth.trosnothgui.ingame.leaderboard import LeaderBoard
from trosnoth.trosnothgui.ingame.statusBar import (
    ZoneProgressBar, FrontLineProgressBar)
from trosnoth.trosnothgui.ingame.gameTimer import GameTimer
from trosnoth.trosnothgui.ingame.sprites import PlayerSprite
from trosnoth.trosnothgui.ingame.universegui import UniverseGUI
from trosnoth.model.map import MapLayout
from trosnoth.model.utils import getZonesInRect, getBlocksInRect
from trosnoth.mumble import mumbleUpdater

ZONE_OFFSETS = {
    'top': (512, 0),
    'btm': (512, 384),
    'fwd': ((1536, 384), (0, 0)),
    'bck': ((1536, 0), (0, 384)),
}
ZONE_SIZE = (2048, 768)

BACKDROP_COLOUR_KEY = (255, 255, 255)

log = logging.getLogger('viewManager')


class ViewManager(framework.Element):
    '''A ViewManager object takes a given universe, and displays a screenfull
    of the current state of the universe on the specified screen object.  This
    class displays only a section of the universe and no other information
    (scores, menu etc.).

    Note: self._focus represents the position that the ViewManager is currently
    looking at.  self.target is what the ViewManager should be trying to look
    at.

    self.target = None - the ViewManager will use its algorithm to follow a
        point of action.
    self.target = (x, y) - the ViewManager will look at the specified point.
    self.target = player - the ViewManager will follow the specified player.
    '''

    # The fastest speed that the viewing position can shift in pixels per sec
    maxSpeed = 1800
    acceleration = 1080

    # How far ahead of the targeted player we should look.
    lengthFromPlayer = 125

    def __init__(self, app, parent, universe, target=None, replay=False):
        '''
        Called upon creation of a ViewManager object.  screen is a pygame
        screen object.  universe is the Universe object to draw.  target is
        either a point, a PlayerSprite object, or None.  If target is None, the
        view manager will follow the action, otherwise it will follow the
        specified point or player.
        '''
        super(ViewManager, self).__init__(app)
        self.universe = universe
        self.parent = parent
        self.replay = replay

        # self._focus represents the point where the viewManager is currently
        # looking.
        self._focus = (
            universe.map.layout.centreX, universe.map.layout.centreY)
        self._oldTargetPt = self._focus
        self.lastUpdateTime = timeNow()
        self.autoFocusInfo = (0, set(), set())
        self.speed = 0          # Speed that the focus is moving.

        self.loadingScreen = None
        self.loadingPlayer = None
        self.backgroundDrawer = BackgroundDrawer(app, universe)
        self.sRect = None

        # Now fill the backdrop with what we're looking at now.
        self.appResized()
        self.setTarget(target)

    def reset(self):
        if self.target is None:
            if self.autoFocusInfo == [0, set(), set()]:
                self._focus = (
                    self.universe.map.layout.centreX,
                    self.universe.map.layout.centreY)

    def stop(self):
        self.backgroundDrawer.stop()

    def tick(self, deltaT):
        if not self.active:
            return
        super(ViewManager, self).tick(deltaT)
        self.updateMumble()

    def updateMumble(self):
        mumbleUpdater.update(
            self.getTargetPlayer(), self.getTargetPoint(), serverId='')

    def appResized(self):
        self.loadingScreen = None
        self.sRect = sRect = pygame.Rect((0, 0), self.app.screenManager.size)
        centre = sRect.center
        if not self.replay:
            settings = self.app.displaySettings
            sRect.width = min(settings.maxViewportWidth, sRect.width)
            sRect.height = min(settings.maxViewportHeight, sRect.height)
        sRect.center = centre

    def setTarget(self, target):
        '''Makes the viewManager's target the specified value.'''
        self.target = target
        if isinstance(self.target, PlayerSprite):
            # Move directly to looking at player.
            self._focus = target.pos
        elif isinstance(self.target, (tuple, list)):
            self.target = self.trimTargetToMap(self.target)
        else:
            countdown, players, zones = self.autoFocusInfo
            if not players and not zones:
                self._oldTargetPt = self._focus

    def getTargetPlayer(self):
        if isinstance(self.target, PlayerSprite):
            return self.target
        return None

    def getTargetPoint(self):
        '''Returns the position of the current target.'''
        if self.target is None:
            return self._focus
        return getattr(self.target, 'pos', self.target)

    def drawLoading(self, screen):
        if self.loadingPlayer is None:
            player = Player(self.universe.universe, 'Loading...', None, '')
            player.motionState = 'ground'
            player.updateState('right', True)
            self.loadingPlayer = PlayerSprite(
                self.app, self.universe, player, timer=timeNow)

        if self.loadingScreen is None:
            self.loadingScreen = pygame.Surface(screen.get_size())
            self.loadingScreen.fill((255, 255, 255))

            font = self.app.screenManager.fonts.mainMenuFont
            colour = self.app.theme.colours.mainMenuColour
            text = font.render(
                self.app, 'Loading...', True, colour, (255, 255, 255))
            r = text.get_rect()
            r.center = self.loadingScreen.get_rect().center
            self.loadingScreen.blit(text, r)

            self.loadingPlayer.rect.midbottom = r.midtop

        if random.random() < 0.03:
            self.loadingPlayer.unit.lookAt(random.random() * 2 * math.pi)

        screen.blit(self.loadingScreen, (0, 0))
        self.loadingPlayer.update()
        screen.fill((255, 255, 255), self.loadingPlayer.rect)
        screen.blit(self.loadingPlayer.image, self.loadingPlayer.rect)

    def draw(self, screen, drawStars=True):
        '''Draws the current state of the universe at the current viewing
        location on the screen.  Does not call pygame.display.flip()'''

        if self.universe.universe.loading:
            self.drawLoading(screen)
            return

        # Update where we're looking at.
        self.updateFocus()

        if self.sRect.topleft != (0, 0):
            screen.fill((0, 0, 0))

        oldClip = screen.get_clip()

        screen.set_clip(self.sRect)
        self.backgroundDrawer.draw(screen, self.sRect, self._focus, drawStars)
        self._drawSprites(screen)
        self.drawOverlay(screen)
        screen.set_clip(oldClip)

    def drawOverlay(self, screen):
        area = self.sRect

        target = self.getTargetPlayer()
        if target is not None:
            physics = target.world.physics
            gunRange = physics.shotLifetime * physics.shotSpeed
            radius = int(gunRange * MAP_TO_SCREEN_SCALE + 0.5)
            pygame.draw.circle(screen, (192, 64, 64), area.center, radius, 1)

    def _drawSprites(self, screen):
        focus = self._focus
        area = self.sRect

        # Go through and update the positions of the players on the screen.
        ntGroup = set()
        visPlayers = set()

        for player in self.universe.iterPlayers():
            self.addSpritesForPlayer(player, visPlayers, ntGroup)

        # Draw the on-screen players and nametags.
        for s in visPlayers:
            s.update()
            screen.blit(s.image, s.rect)
        for s in ntGroup:
            screen.blit(s.image, s.rect)

        def drawSprite(sprite, pos=None):
            # Calculate the position of the sprite.
            if pos is None:
                pos = sprite.pos
            sprite.rect.center = mapPosToScreen(pos, focus, area)
            if sprite.rect.colliderect(area):
                sprite.update()
                screen.blit(sprite.image, sprite.rect)

        # Draw the shots.
        for shot in self.universe.iterShots():
            drawSprite(shot)

        for star in self.universe.iterCollectableStars():
            drawSprite(star)

        try:
            # Draw the grenades.
            for grenade in self.universe.iterGrenades():
                drawSprite(grenade)
        except Exception, e:
            log.exception(str(e))

        for sprite in self.universe.iterExtras():
            drawSprite(sprite)

        if __debug__ and globaldebug.enabled:
            if globaldebug.showSpriteCircles:
                for pos, radius in globaldebug.getSpriteCircles():
                    screenPos = mapPosToScreen(pos, focus, area)
                    pygame.draw.circle(
                        screen, (255, 255, 0), screenPos, radius, 2)

            for sprite in visPlayers:
                sprite.player.onOverlayDebugHook(self, screen, sprite)

    def addSpritesForPlayer(self, player, visPlayers, ntGroup):
        focus = self._focus
        area = self.sRect

        targetPlayer = self.getTargetPlayer()
        showPlayer = (
            not player.invisible or targetPlayer is None
            or player.isFriendsWith(targetPlayer.player))

        if showPlayer:
            # Calculate the position of the player.
            if player is targetPlayer:
                player.rect.center = area.center
            else:
                player.rect.center = mapPosToScreen(player.pos, focus, area)

            # Check if this player needs its nametag shown.
            if player.rect.colliderect(area):
                visPlayers.add(player)

                if ntGroup is None:
                    return

                if (
                        player.unit.upgrade
                        and player.unit.upgrade.upgradeType == 'b'):
                    player.countdown.update()
                    player.countdown.rect.midbottom = player.rect.midtop
                    ntGroup.add(player.countdown)

                lastPoint = player.rect.midbottom

                if player.unit.hasProtectiveUpgrade():
                    shieldBar = player.shieldBar
                    shieldBar.setHealth(
                        player.unit.upgrade.protections,
                        player.unit.upgrade.maxProtections)
                    if shieldBar.visible:
                        ntGroup.add(shieldBar)
                        shieldBar.rect.midtop = lastPoint
                        lastPoint = shieldBar.rect.midbottom

                healthBar = player.healthBar
                healthBar.setHealth(
                    player.unit.health,
                    player.unit.world.physics.playerRespawnHealth)
                if healthBar.visible:
                    ntGroup.add(healthBar)
                    healthBar.rect.midtop = lastPoint
                    lastPoint = healthBar.rect.midbottom

                player.nametag.rect.midtop = lastPoint

                # Check that entire nametag's on screen.
                if player.nametag.rect.left < area.left:
                    player.nametag.rect.left = area.left
                elif player.nametag.rect.right > area.right:
                    player.nametag.rect.right = area.right
                if player.nametag.rect.top < area.top:
                    player.nametag.rect.top = area.top
                elif player.nametag.rect.bottom > area.bottom:
                    player.nametag.rect.bottom = area.bottom
                ntGroup.add(player.nametag)

                # Place the star rectangle below the nametag.
                mx, my = player.nametag.rect.midbottom
                player.starTally.setStars(player.stars)
                player.starTally.rect.midtop = (mx, my - 5)
                ntGroup.add(player.starTally)

    def updateFocus(self):
        '''Updates the location that the ViewManager is focused on.  First
        calculates where it would ideally be focused, then moves the focus
        towards that point. The focus cannot move faster than self.maxSpeed
        pix/s, and will only accelerate or decelerate at self.acceleration
        pix/s/s. This method returns the negative of the amount scrolled by.
        This is useful for moving the backdrop by the right amount.
        '''

        # Calculate where we should be looking at.
        if isinstance(self.target, PlayerSprite):
            # Take into account where the player's looking.
            targetPt = self.target.pos

            # If the player no longer exists, look wherever we want.
            if not self.universe.hasPlayer(self.target.player):
                self.setTarget(None)
        elif isinstance(self.target, (tuple, list)):
            targetPt = self.target
        else:
            targetPt = self.followAction()

        # Calculate time that's passed.
        now = timeNow()
        deltaT = now - self.lastUpdateTime
        self.lastUpdateTime = now

        # Calculate distance to target.
        self._oldTargetPt = targetPt
        sTarget = sum(
            (targetPt[i] - self._focus[i]) ** 2 for i in (0, 1)) ** 0.5

        if sTarget == 0:
            return (0, 0)

        if self.target is not None:
            s = sTarget
        else:
            # Calculate the maximum velocity that will result in deceleration
            # to reach target. This is based on v**2 = u**2 + 2as
            vDecel = (2. * self.acceleration * sTarget) ** 0.5

            # Actual velocity is limited by this and maximum velocity.
            self.speed = min(
                self.maxSpeed, vDecel, self.speed + self.acceleration * deltaT)

            # Distance travelled should never overshoot the target.
            s = min(sTarget, self.speed * deltaT)

        # How far does the backdrop need to move by?
        #  (This will be negative what the focus moves by.)
        deltaBackdrop = tuple(
            -s * (targetPt[i] - self._focus[i]) / sTarget
            for i in (0, 1))

        # Calculate the new focus.
        self._focus = tuple(
            round(self._focus[i] - deltaBackdrop[i], 0) for i in (0, 1))

    def getZoneAtPoint(self, pt):
        x, y = screenToMapPos(pt, self._focus, self.sRect)

        i, j = MapLayout.getMapBlockIndices(x, y)
        try:
            return self.universe.map.zoneBlocks[i][j].getZoneAtPoint(x, y)
        except IndexError:
            return None

    def followAction(self):
        # Follow the action.
        countdown, players, zones = self.autoFocusInfo
        targetPt = tuple(self._focus)

        if self.universe.getPlayerCount() == 0:
            # No players anywhere. No change.
            self.autoFocusInfo = (0, set(), set())
            return targetPt

        # First check for non-existent players.
        for p in list(players):
            if not self.universe.hasPlayer(p):
                players.remove(p)

        # Every 10 iterations recheck for players that have entered
        # view area.
        r = pygame.Rect(self.sRect)
        r.width //= MAP_TO_SCREEN_SCALE
        r.height //= MAP_TO_SCREEN_SCALE
        r.center = self._oldTargetPt
        if countdown <= 0:
            players.clear()
            for p in self.universe.iterPlayers():
                if r.collidepoint(p.pos):
                    players.add(p)

            zones.clear()
            for z in self.universe.map.zones:
                if r.collidepoint(z.defn.pos):
                    if any(t != z.owner for t in z.teamsAbleToTag()):
                        zones.add(z)
            countdown = 10
        else:
            # Keep track of which players are still visible.
            for p in list(players):
                if not r.collidepoint(p.pos):
                    players.remove(p)
            countdown -= 1

        if len(players) + len(zones) <= 1:
            # Nothing interesting in view. Look for action.
            maxP = 0
            curZone = None
            for z in self.universe.zones:
                count = len(self.universe.getPlayersInZone(z))
                if any(t != z.owner for t in z.teamsAbleToTag()):
                    count += 2
                if count > maxP:
                    maxP = count
                    curZone = z
            if curZone is None:
                players = set()
            else:
                players = set(self.universe.getPlayersInZone(curZone))

            countdown = 20

        # Look at centre-of-range of these players.
        if players or zones:
            interestingPoints = []
            for p in players:
                interestingPoints.append(p.pos)
            for z in zones:
                interestingPoints.append(z.defn.pos)
            minPos = [
                min(pt[i] for pt in interestingPoints) for i in (0, 1)]
            maxPos = [
                max(pt[i] for pt in interestingPoints) for i in (0, 1)]
            targetPt = [0.5 * (minPos[i] + maxPos[i]) for i in (0, 1)]

        # No need to ever look beyond the boundary of the map
        targetPt = self.trimTargetToMap(targetPt)
        r.center = targetPt

        self.autoFocusInfo = (countdown, players, zones)
        return targetPt

    def trimTargetToMap(self, targetPt):
        # No need to ever look beyond the boundary of the map
        mapRect = pygame.Rect((0, 0), self.universe.map.layout.worldSize)
        r = pygame.Rect(self.sRect)
        r.width //= MAP_TO_SCREEN_SCALE
        r.height //= MAP_TO_SCREEN_SCALE
        r.center = targetPt
        if r.width > mapRect.width:
            r.centerx = mapRect.centerx
        else:
            r.right = min(r.right, mapRect.right)
            r.left = max(r.left, mapRect.left)
        if r.height > mapRect.height:
            r.centery = mapRect.centery
        else:
            r.bottom = min(r.bottom, mapRect.bottom)
            r.top = max(r.top, mapRect.top)
        return r.center


class BackgroundDrawer(object):
    def __init__(self, app, universe):
        self.app = app
        self.scenery = Scenery(app, universe)
        self.sBackgrounds = SolidBackgrounds(app, universe)
        self.orbs = OrbDrawer(app, universe)
        self.debugs = DebugDrawer(app, universe)

        app.displaySettings.onDetailLevelChanged.addListener(
            self.detailLevelChanged)

    def stop(self):
        self.app.displaySettings.onDetailLevelChanged.removeListener(
            self.detailLevelChanged)

    def detailLevelChanged(self):
        self.sBackgrounds.bkgCache.clear()

    def draw(self, screen, sRect, focus, drawStars=True):
        if drawStars:
            self.scenery.draw(screen, sRect, focus)
        self.sBackgrounds.draw(screen, sRect, focus)
        self.orbs.draw(screen, sRect, focus)
        self.debugs.draw(screen, sRect, focus)


class Scenery(object):
    def __init__(self, app, universe, distance=3):
        self.app = app
        self.universe = universe
        self.image = app.theme.sprites.scenery
        self.scale = 1. / distance

    def draw(self, screen, area, focus):
        worldRect = viewRectToMap(focus, area)

        regions = []
        for block in getBlocksInRect(self.universe, worldRect):
            bd = block.defn
            pos = mapPosToScreen(bd.pos, focus, area)
            if bd.kind in ('top', 'btm'):
                if bd.zone is None:
                    regions.append(pygame.Rect(pos, BODY_BLOCK_SCREEN_SIZE))
                    continue
            elif bd.zone1 is None or bd.zone2 is None:
                regions.append(pygame.Rect(pos, INTERFACE_BLOCK_SCREEN_SIZE))
                continue

        x0, y0 = mapPosToScreen((0, 0), focus, area)
        if area.top < y0:
            r = pygame.Rect(area)
            r.bottom = y0
            regions.append(r)
        if area.left < x0:
            r = pygame.Rect(area)
            r.right = x0
            regions.append(r)

        x1, y1 = mapPosToScreen(
            self.universe.map.layout.worldSize, focus, area)
        if area.bottom > y1:
            r = pygame.Rect(area)
            r.top = y1
            regions.append(r)
        if area.right > x1:
            r = pygame.Rect(area)
            r.left = x1
            regions.append(r)

        clip = screen.get_clip()
        for region in regions:
            region = region.clip(clip)
            screen.set_clip(region)
            self.drawRegion(screen, region, worldRect.topleft)
        screen.set_clip(clip)

    def drawRegion(self, screen, area, focus):
        if not self.app.displaySettings.paralaxBackgrounds:
            screen.fill(BLOCK_BACKGROUND_COLOURKEY, area)
            return

        w, h = self.image.get_size()
        x = area.left - (int(round(focus[0] * self.scale + area.left)) % w)
        y0 = y = area.top - (int(round(focus[1] * self.scale + area.top)) % h)

        while x < area.right:
            while y < area.bottom:
                screen.blit(self.image, (x, y))
                y += h
            x += w
            y = y0


class OrbDrawer(object):
    def __init__(self, app, world):
        self.app = app
        self.universe = world

    def draw(self, screen, area, focus):
        worldRect = viewRectToMap(focus, area)

        for zone in getZonesInRect(self.universe, worldRect):
            pic = self.app.theme.sprites.bigZoneLetter(zone.defn.label)
            r = pic.get_rect()
            r.center = mapPosToScreen(zone.defn.pos, focus, area)
            screen.blit(pic, r)

            if (self.universe.universe.uiOptions.showNets and zone.defn in
                    self.universe.map.layout.trosballTargetZones):
                pic = self.app.theme.sprites.netOrb()
            else:
                pic = self.app.theme.sprites.orb(zone.owner)
            r = pic.get_rect()
            r.center = mapPosToScreen(zone.defn.pos, focus, area)
            screen.blit(pic, r)


class DebugDrawer(object):
    def __init__(self, app, world):
        self.app = app
        self.universe = world

    def draw(self, screen, area, focus):
        if not self.app.displaySettings.showObstacles:
            return

        from trosnoth.model.obstacles import Obstacle, Corner

        player = self.universe.universe.getPlayer(globaldebug.localPlayerId)
        attachedObstacle = player.attachedObstacle if player else None

        worldRect = viewRectToMap(focus, area)
        for block in getBlocksInRect(self.universe, worldRect):
            for obs in block.defn.obstacles:
                if isinstance(obs, Obstacle):
                    pt1 = mapPosToScreen(obs.pt1, focus, area)
                    pt2 = mapPosToScreen(obs.pt2, focus, area)
                    c = (0, 255, 0) if obs is attachedObstacle else (255, 0, 0)
                    pygame.draw.line(screen, c, pt1, pt2, 2)
                elif isinstance(obs, Corner):
                    pt1 = mapPosToScreen(
                        [obs.pt[i] - obs.offset[i] * 10 for i in (0, 1)],
                        focus, area)
                    pt2 = mapPosToScreen([
                        obs.pt[i] - (obs.offset[i] + obs.delta[i]) * 10
                        for i in (0, 1)], focus, area)
                    c = (0 if obs is attachedObstacle else 255, 255, 0)
                    pygame.draw.line(screen, c, pt1, pt2, 2)
                    pt2 = (int(pt2[0]), int(pt2[1]))
                    pygame.draw.circle(screen, c, pt2, 3, 0)


class SolidBackgrounds(object):
    def __init__(self, app, universe):
        self.app = app
        self.universe = universe
        self.bkgCache = BackgroundCache(app, universe)

    def draw(self, screen, area, focus):
        if self.universe.universe.uiOptions.shiftingFrontLine:
            self.drawShiftingBackground(screen, area, focus)
        else:
            self.drawStandardBackground(screen, area, focus)

    def drawStandardBackground(self, screen, area, focus):
        worldRect = viewRectToMap(focus, area)
        for block in getBlocksInRect(self.universe, worldRect):
            pic = self.bkgCache.get(block)
            if pic is not None:
                screen.blit(pic, mapPosToScreen(block.defn.pos, focus, area))

    def drawShiftingBackground(self, screen, area, focus):
        worldRect = viewRectToMap(focus, area)
        uiOptions = self.universe.universe.uiOptions
        trosballLocation = uiOptions.frontLineGetter()

        for block in getBlocksInRect(self.universe, worldRect):
            blueBlock = self.bkgCache.getForTeam(0, block)
            redBlock = self.bkgCache.getForTeam(1, block)
            if blueBlock is None or redBlock is None:
                continue

            blockHorizontalPosition = block.defn.pos[0]
            relativeLocation = trosballLocation - blockHorizontalPosition
            blockWidth = block.defn._getWidth()
            relativeLocation = max(0, min(blockWidth, relativeLocation))

            x = int(relativeLocation * MAP_TO_SCREEN_SCALE + 0.5)
            topleft = mapPosToScreen(block.defn.pos, focus, area)

            r = blueBlock.get_rect()
            r.width = x
            screen.blit(blueBlock, topleft, r)

            r = redBlock.get_rect()
            r.left = x
            screen.blit(redBlock, (topleft[0] + x, topleft[1]), r)


class BackgroundCache(object):
    def __init__(self, app, universe, capacity=30):
        self.app = app
        self.universe = universe
        self.capacity = capacity
        self.cache = {}
        self.order = []

    def clear(self):
        self.cache = {}
        self.order = []

    def getForTeam(self, teamId, block):
        backgroundPicTeam = self.app.theme.sprites.getFilledBlockBackground(
            block, self.universe.teams[teamId])
        if backgroundPicTeam is None:
            return None
        return self._getForegroundOnBackground(backgroundPicTeam, block)

    def get(self, block):
        backgroundPic = self.app.theme.sprites.blockBackground(block)
        return self._getForegroundOnBackground(backgroundPic, block)

    def _getForegroundOnBackground(self, backgroundPic, block):
        if block.defn.graphics is not None:
            foregroundPic = block.defn.graphics.getGraphic(self.app)
        else:
            foregroundPic = None

        if (backgroundPic, foregroundPic) in self.cache:
            self.order.remove((backgroundPic, foregroundPic))
            self.order.insert(0, (backgroundPic, foregroundPic))
            return self.cache[backgroundPic, foregroundPic]

        pic = self._makePic(backgroundPic, foregroundPic)
        self.cache[backgroundPic, foregroundPic] = pic
        self.order.insert(0, (backgroundPic, foregroundPic))
        if len(self.order) > self.capacity:
            del self.cache[self.order.pop(-1)]
        return pic

    def _makePic(self, backgroundPic, foregroundPic):
        if backgroundPic is None:
            return foregroundPic
        if foregroundPic is None:
            return backgroundPic
        pic = backgroundPic.copy()
        pic.blit(foregroundPic, (0, 0))
        return pic


class GameViewer(framework.CompoundElement):
    '''The gameviewer comprises a viewmanager and a minimap, which can be
    switched on or off.'''

    zoneBarHeight = 25

    def __init__(self, app, gameInterface, game, replay):
        super(GameViewer, self).__init__(app)
        self.replay = replay
        self.interface = gameInterface
        self.game = game
        self.world = game.world
        self.worldgui = UniverseGUI(app, self, self.world)
        self.app = app

        self.viewManager = ViewManager(
            self.app, self, self.worldgui, replay=replay)

        self.timerBar = GameTimer(app, game)

        self.miniMap = None
        self.leaderboard = None
        self.zoneBar = None
        self.makeWidgets()

        self.elements = [self.viewManager]
        self._screenSize = tuple(app.screenManager.size)

        self.teamsDisrupted = set()

        self.storedState = None

        self.toggleInterface()
        self.toggleLeaderBoard()

    def stop(self):
        self.viewManager.stop()
        self.worldgui.stop()

    def getZoneAtPoint(self, pos):
        '''
        Returns the zone at the given screen position. This may be on the
        minimap or the main view.
        '''
        zone = self.miniMap.getZoneAtPoint(pos)
        if zone is None:
            zone = self.viewManager.getZoneAtPoint(pos)
        return zone

    def resizeIfNeeded(self):
        '''
        Checks whether the application has resized and adjusts accordingly.
        '''
        if self._screenSize == self.app.screenManager.size:
            return
        self._screenSize = tuple(self.app.screenManager.size)

        self.viewManager.appResized()
        # Recreate the minimap.
        self.reset()

    def reset(self):
        showHUD = self.miniMap is not None and self.miniMap in self.elements
        showLeader = (
            self.leaderboard is not None and self.leaderboard in self.elements)
        self.makeWidgets()
        if showHUD:
            self.toggleInterface()
        if showLeader:
            self.toggleLeaderBoard()
        self.viewManager.reset()

    def makeWidgets(self):
        self.miniMap = MiniMap(self.app, 20, self.worldgui, self.viewManager)
        if self.world.uiOptions.shiftingFrontLine:
            self.zoneBar = FrontLineProgressBar(self.app, self.world, self)
            self.leaderboard = None
        else:
            self.zoneBar = ZoneProgressBar(self.app, self.world, self)

        self.leaderboard = LeaderBoard(self.app, self.game, self)

        self.elements = [self.viewManager]

    def setTarget(self, target):
        'Target should be a player, a point, or None.'
        self.viewManager.setTarget(target)

    def tick(self, deltaT):
        if not self.active:
            return
        self.resizeIfNeeded()
        self.worldgui.setTweenFraction(self.app.tweener.uiTick(deltaT))
        if self.world.state != self.storedState:
            self.storedState = self.world.state
            self.reset()

        target = self.viewManager.target
        if isinstance(target, PlayerSprite) and target.isMinimapDisrupted:
            self.miniMap.disrupted()
            self.zoneBar.disrupt = True
        else:
            self.miniMap.endDisruption()
            self.zoneBar.disrupt = False

        super(GameViewer, self).tick(deltaT)

    def toggleInterface(self):
        if self.miniMap in self.elements:
            self.elements.remove(self.zoneBar)
            self.elements.remove(self.timerBar)
            self.elements.remove(self.miniMap)
        else:
            self.elements.append(self.zoneBar)
            self.elements.append(self.timerBar)
            self.elements.append(self.miniMap)
            self.timerBar.syncTimer()

    def toggleLeaderBoard(self):
        if self.leaderboard is not None and self.leaderboard in self.elements:
            self.elements.remove(self.leaderboard)
        else:
            self.elements.append(self.leaderboard)



import pygame

from trosnoth.gui.framework.elements import TextElement, PictureElement
from trosnoth.gui.common import (Area, FullScreenAttachedPoint, Location,
        ScaledSize)
import trosnoth.gui.framework.framework as framework
from trosnoth.gamerecording.achievementlist import availableAchievements
from trosnoth.utils.twist import WeakLoopingCall

class AchievementBox(framework.CompoundElement):
    achievementDefs = availableAchievements
    def __init__(self, app, player, achievementId):
        super(AchievementBox, self).__init__(app)
        self.app = app
        self.player = player

        self.achievements = [achievementId]

        self.width = 453
        self.height = 75

        self._setColours()

        self.area = Area(FullScreenAttachedPoint(ScaledSize(0, -100),
                'midbottom'), ScaledSize(self.width, self.height), 'midbottom')

        self.smlBox = Area(FullScreenAttachedPoint(
                ScaledSize(-self.width / 2 + 6, -104), 'midbottom'),
                ScaledSize(66, 66), 'bottomleft')

        self.titleText = TextElement(self.app, "ACHIEVEMENT UNLOCKED!",
                self.fonts.achievementTitleFont,
                Location(FullScreenAttachedPoint(ScaledSize(73 / 2, -100 -
                self.height + 10), 'midbottom'), 'midtop'), self.borderColour)

        self.nameText = TextElement(self.app,
                self.achievementDefs.getAchievementDetails(achievementId)[0],
                self.fonts.achievementNameFont,
                Location(FullScreenAttachedPoint(ScaledSize(73 / 2, -100 - 13),
                'midbottom'), 'midbottom'), self.colours.black)

        self.elements = [self.titleText, self.nameText]
        self._updateImage()

        self.cycler = WeakLoopingCall(self, 'cycleAchievements')
        self.cycler.start(5.0, now = False)

    def addAchievement(self, achievementId):
        self.achievements.append(achievementId)
        self.titleText.setText("%d ACHIEVEMENTS UNLOCKED!" %
                len(self.achievements))

    def _updateImage(self):
        try:
            filepath = self.app.theme.getPath('achievements', '%s.png' %
                    self.achievements[0])
        except IOError:
            filepath = self.app.theme.getPath('achievements', 'default.png')

        image = pygame.image.load(filepath).convert()
        image = pygame.transform.smoothscale(image,
                ScaledSize(64, 64).getSize(self.app))

        if type(self.elements[-1]) == PictureElement:
            self.elements.pop()

        self.image = PictureElement(self.app, image,
                Location(FullScreenAttachedPoint(
                ScaledSize(-self.width / 2 + 7, -105), 'midbottom'),
                'bottomleft'))

        self.elements.append(self.image)

    def cycleAchievements(self):
        del self.achievements[0]

        if len(self.achievements) == 0:
            self.cycler.stop()
            return
        elif len(self.achievements) == 1:
            self.titleText.setText("ACHIEVEMENT UNLOCKED!")
        else:
            self.titleText.setText("%d ACHIEVEMENTS UNLOCKED!" %
                    len(self.achievements))

        self.nameText.setText(self.achievementDefs.getAchievementDetails(
                self.achievements[0])[0])
        self._updateImage()

    def _setColours(self):
        self.colours = self.app.theme.colours
        self.fonts = self.app.screenManager.fonts

        if self.player.teamId == 'A':
            self.borderColour = self.colours.achvBlueBorder
            self.bgColour = self.colours.achvBlueBackground
        else:
            self.borderColour = self.colours.achvRedBorder
            self.bgColour = self.colours.achvRedBackground

    def _getRect(self, area):
        return area.getRect(self.app)

    def draw(self, surface):
        mainRect = self._getRect(self.area)
        boxRect = self._getRect(self.smlBox)

        surface.fill(self.bgColour, mainRect)
        pygame.draw.rect(surface, self.borderColour, mainRect, 2)

        surface.fill(self.colours.white, boxRect)
        pygame.draw.rect(surface, self.borderColour, boxRect, 1)

        super(AchievementBox, self).draw(surface)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('trosnoth', '0006_auto_20160209_2008'),
    ]

    operations = [
        migrations.AddField(
            model_name='trosnothserversettings',
            name='allowRemoteGameRegistration',
            field=models.BooleanField(default=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='achievementprogress',
            name='progress',
            field=models.IntegerField(default=0),
            preserve_default=True,
        ),
    ]



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.utils.event import Event

class MenuManager(object):
    def __init__(self, defaultMenu=None):
        self.defaultMenu = defaultMenu
        self.menu = defaultMenu
        self.menuStack = []
        self.onShowMenu = Event()   # (menu)

    def setDefaultMenu(self, defaultMenu):
        default = len(self.menuStack) == 0 and self.menu is self.defaultMenu
        self.defaultMenu = defaultMenu

        if default or self.menu is None:
            self.menu = defaultMenu
            self.onShowMenu.execute(defaultMenu)

    def reset(self):
        self.menu = self.defaultMenu
        self.menuStack = []

        self.onShowMenu.execute(self.defaultMenu)

    def cancel(self):
        if len(self.menuStack) > 0:
            self.menu = self.menuStack.pop()
        else:
            self.menu = self.defaultMenu

        self.onShowMenu.execute(self.menu)

    def showMenu(self, menu):
        self.menuStack.append(self.menu)
        self.menu = menu

        self.onShowMenu.execute(menu)

class MenuItem(object):
    '''
    @param name     the caption of the menu item
    @param action   a callable for when the item is selected
    '''
    def __init__(self, name, action=None, listener=None, enabled=True):
        self.name = name
        self.onClick = Event()      # (MenuItem)
        self.onAction = Event()     # (action)
        self.action = action
        self.enabled = enabled
        if listener is not None:
            self.onClick.addListener(listener)

    def execute(self):
        self.onClick.execute(self)
        if self.action is not None:
            self.onAction.execute(self.action)

class Menu(object):
    '''
    @param name     the caption of this menu
    @param items    a sequence of MenuItem objects to be displayed in this menu
    '''
    def __init__(self, name, items=(), listener=None):
        self.name = name
        self.items = list(items)

        self.onAction = Event()     # (action)
        if listener is not None:
            self.onAction.addListener(listener)
        for item in items:
            item.onAction.addListener(self.onAction.execute)

    def popItem(self, index):
        item = self.items.pop(index)
        item.onAction.removeListener(self.onAction.execute)
        return item

    def removeItem(self, item):
        self.popItem(self.items.index(item))

    def insertItem(self, pos, item):
        self.items.insert(pos, item)
        item.onAction.addListener(self.onAction.execute)

    def appendItem(self, item):
        self.insertItem(len(self.items), item)



from trosnoth.gui.screenManager.screenManager import ScreenManager
from trosnoth.gui.errors import MultiWindowException

class WindowManager(ScreenManager):
    '''Handles all the dialog boxes and the main screen'''

    def __init__(self, app, element, size, settings, caption):
        super(WindowManager, self).__init__(app, size, settings, caption)
        self.boxes = []

    def createInterface(self, element):
        super(WindowManager,self).createInterface(element)
        self._fixElements()

    def showDialog(self, dialog):
        self.boxes.insert(len(self.elements)-1, dialog)
        self._fixElements()

    def closeDialog(self, dialog):
        if dialog not in self.boxes:
            raise MultiWindowException(
                    "Attempt to close a dialog that was not open")
        self.boxes.remove(dialog)
        self._fixElements()

    def dialogFocus(self, dialog):
        index = self.boxes.index(dialog)
        self.boxes = (self.boxes[:index] + self.boxes[index+1:] +
                [self.boxes[index]])
        self._fixElements()

    def _fixElements(self):
        self.elements = [self.interface]+ self.boxes
        if self.pointer is not None:
            self.elements.append(self.pointer)
        self.elements.append(self.terminator)



'''
themes.py
This module defines the interface to the various different themes.
'''

import functools
import logging
import os
from random import randrange

import pygame
import pygame.transform

from trosnoth import data
import trosnoth.data.themes     # noqa
from trosnoth.const import BODY_BLOCK_SCREEN_SIZE, INTERFACE_BLOCK_SCREEN_SIZE
from trosnoth.gui.fonts.font import Font, ScaledFont
from trosnoth.gui.framework.basics import SingleImage, Animation
from trosnoth.trosnothgui.common import setAlpha
from trosnoth.utils.unrepr import unrepr
from trosnoth.model.upgrades import Upgrade
from trosnoth.utils.utils import timeNow

BLOCK_BACKGROUND_COLOURKEY = (224, 192, 224)
x1 = INTERFACE_BLOCK_SCREEN_SIZE[0]
x2 = x1 + BODY_BLOCK_SCREEN_SIZE[0]
y = BODY_BLOCK_SCREEN_SIZE[1]
BLOCK_OFFSETS = {
    'top': (-x1, 0),
    'btm': (-x1, -y),
    'fwd': ((-x2, -y), (0, 0)),
    'bck': ((-x2, 0), (0, -y)),
}
del x1, x2, y

log = logging.getLogger('themes')


def setToRed(surface):
    '''Inverts the colors of a pygame Screen'''
    surface.lock()

    for x in range(surface.get_width()):
        for y in range(surface.get_height()):
            r, g, b, a = surface.get_at((x, y))
            greyscale = r * 0.298 + g * 0.587 + b * 0.114
            surface.set_at((x, y), (greyscale, 0, 0, a))

    surface.unlock()


def teamColour(colourId):
    def colourFunction(self, team):
        return self.getTeamColour(team, colourId)
    return colourFunction


class ThemeColours(object):
    def getTeamColour(self, team, colourId):
        if team is None:
            teamNum = 0
        else:
            teamNum = ord(team.id) - 64
        return getattr(
            self, 'team%d%s' % (teamNum, colourId),
            getattr(self, 'team0%s' % (colourId,)))

    sysMessageColour = teamColour('msg')
    chatColour = teamColour('chat')
    miniMapTrosballTargetColour = teamColour('Mn_trosball')
    miniMapDarkZoneColour = teamColour('Mn_zone')
    miniMapLightZoneColour = teamColour('Mn_mk')
    miniMapPlayerColour = teamColour('Mn_pl')
    miniMapGhostColour = teamColour('Mn_gh')
    miniMapZoneHighlight = teamColour('Mn_highlight')
    leaderboard = teamColour('Leaderboard')


def cachedProperty(fn):
    @functools.wraps(fn)
    def spriteFunction(self):
        try:
            return self._store[fn]
        except KeyError:
            self._store[fn] = fn(self)
            return self._store[fn]
    return property(spriteFunction)


def cached(fn):
    @functools.wraps(fn)
    def spriteFunction(self, *args):
        try:
            return self._store[fn, args]
        except KeyError:
            self._store[fn, args] = fn(self, *args)
            return self._store[fn, args]
    return spriteFunction


def image(path, **kwargs):
    def imageFunction(self):
        return self.theme.loadSprite(path, sprites=self, **kwargs)
    imageFunction.__name__ = path
    return cachedProperty(imageFunction)


def images(paths, **kwargs):
    def imageFunction(self):
        return self.theme.loadSprites(paths, sprites=self, **kwargs)
    imageFunction.__name__ = ', '.join(paths)
    return cachedProperty(imageFunction)


def wrappedImage(path, **kwargs):
    def imageFunction(self):
        return SingleImage(self.theme.loadSprite(path, sprites=self, **kwargs))
    imageFunction.__name__ = path
    return cachedProperty(imageFunction)


def getTeamId(team):
    if team is None:
        return '\x00'
    return team.id


class ThemeSpritesBase(object):
    def __init__(self, theme):
        self.theme = theme
        self._store = {}

    def clearCache(self):
        self._store = {}

    smallStar = image('smallstar.png')
    star = image('star.png')
    grenade = image('grenade.png')
    neutralOrb = image('greyOrb.png')
    scenery = image('scenery.png', colourkey=None)
    shoxGun3 = image('shoxgun-3.png')

    gunIcon = wrappedImage('gun.png')
    playerBody = wrappedImage('backbone.png', borders=(0, 0, 1))
    elephant = wrappedImage('elephant.png')
    christmasHat = wrappedImage('christmas.png')
    turretBase = wrappedImage('legs-turret.png')
    playerStanding = wrappedImage('legs-s.png')
    playerJumping = wrappedImage('legs-r3.png')
    ninjaHead = wrappedImage('ninjahead.png')
    blocker1 = image('blocker-1.png')
    blocker2 = image('blocker-2.png')
    bomber1 = image('bomber-1.png')
    bomber2 = image('bomber-2.png')
    jammingHatImage = image('jamminghat.png')
    jammingHatSparks = images([
        'jamminghat-spark1.png', 'jamminghat-spark2.png',
        'jamminghat-spark3.png'])

    runningLegs = images([
        'legs-r1.png', 'legs-r2.png', 'legs-r3.png', 'legs-r4.png'])
    backwardsLegs = images([
        'legs-w0.png', 'legs-w1.png', 'legs-w2.png', 'legs-w3.png'])

    shieldImages = images([
        'shieldImage1.png', 'shieldImage2.png', 'shieldImage3.png',
        'shieldImage2.png'])

    @cached
    def zoneHighlight(self, team, scale):
        colour = self.theme.colours.miniMapZoneHighlight(team)
        size = int(300. / scale + 0.5)
        result = pygame.Surface((2 * size, 2 * size))
        result.fill((0, 0, 0))
        result.set_colorkey((0, 0, 0))
        pygame.draw.circle(result, colour, (size, size), size)
        result.set_alpha(64)
        return result

    @cached
    def bigZoneLetter(self, letter):
        font = self.theme.app.screenManager.fonts.bigZoneFont
        result = font.render(
            self.theme.app, letter, True, self.theme.colours.bigZoneLetter)
        setAlpha(result, 128)
        return result

    @cached
    def ghostAnimation(self, team):
        if team is not None:
            frames = self.theme.loadTeamSprites(
                ['ghost1', 'ghost2', 'ghost3', 'ghost4', 'ghost3', 'ghost2'],
                team.id, self)
        else:
            lts = functools.partial(self.theme.loadTeamSprite, sprites=self)
            frames = [
                lts('ghost1', 'A'), lts('ghost2', 'B'), lts('ghost3', 'A'),
                lts('ghost4', 'B'), lts('ghost3', 'A'), lts('ghost2', 'B')]
        return [Animation(0.25, timeNow, *frames)]

    @cached
    def ghostIcon(self, team):
        return SingleImage(self.theme.loadTeamSprite(
            'ghost1', getTeamId(team), self))

    @cached
    def trosballAnimation(self):
        frame0 = self.theme.loadSprite('trosball.png', sprites=self)
        scale = 25. / max(frame0.get_size())
        frames = []
        for theta in xrange(0, 360, 15):
            frames.append(pygame.transform.rotozoom(frame0, -theta, scale))
        return Animation(0.05, timeNow, *frames)

    @cached
    def trosballWarningAnimation(self):
        frame0 = self.theme.loadSprite('trosball.png', sprites=self)
        scale = 25. / max(frame0.get_size())
        frames = []
        for theta in xrange(0, 360, 15):
            surface = pygame.transform.rotozoom(frame0, -theta, scale)
            # Every 90 degrees, invert the colours
            if (theta / 45) % 2 < 1:
                setToRed(surface)
            frames.append(surface)
        return Animation(0.05, timeNow, *frames)

    def explosion(self):
        return Animation(
            0.07, timeNow, *(self.explosionFrame(i) for i in xrange(4)))

    def shoxwaveExplosion(self):
        return Animation(
            0.05, timeNow, *(self.shoxwaveFrame(i) for i in xrange(3)))

    def trosballExplosion(self):
        return Animation(0.07, timeNow, self.explosionFrame(0))

    def jammingHat(self):
        images = []
        for spark in self.jammingHatSparks:
            images.append(spark)
            images.extend([self.jammingHatImage] * randrange(5, 15))
        return Animation(0.05, timeNow, *images)

    @cached
    def explosionFrame(self, frame):
        return self.theme.loadSprite(
            'explosion%d.png' % (frame + 1,), sprites=self)

    @cached
    def shoxwaveFrame(self, frame):
        return self.theme.loadSprite(
            'shoxwave%d.png' % (frame + 1,), sprites=self)

    @cached
    def playerHolding(self, team):
        return SingleImage(self.theme.loadTeamSprite(
            'hold', getTeamId(team), self, borders=(0, 0, 1)))

    @cached
    def machineGunImages(self, team):
        return self.theme.loadTeamSprites([
            'machinegun-0', 'machinegun-1', 'machinegun-2',
            'machinegun-3', 'machinegun-4', 'machinegun-5',
            'machinegun-6', 'machinegun-7', 'machinegun-8'
        ], getTeamId(team), self, borders=(0, 0, 1))

    @cached
    def ricoGunImages(self, team):
        return self.theme.loadTeamSprites([
            'ricogun-0', 'ricogun-1', 'ricogun-2',
            'ricogun-3', 'ricogun-4', 'ricogun-5',
            'ricogun-6', 'ricogun-7', 'ricogun-8'
        ], getTeamId(team), self, borders=(0, 0, 1))

    @cached
    def gunImages(self, team):
        return self.theme.loadTeamSprites([
            'gun-0', 'gun-1', 'gun-2',
            'gun-3', 'gun-4', 'gun-5',
            'gun-6', 'gun-7', 'gun-8'
        ], getTeamId(team), self, borders=(0, 0, 1))

    @cached
    def playerHead(self, team, bot):
        teamId = getTeamId(team)
        if teamId == 'A':
            path = 'blue'
        elif teamId == 'B':
            path = 'red'
        elif teamId == '\x00':
            path = 'white'
        else:
            path = 'team' + teamId + 'Head'
        if bot:
            path += '-bot'
        return SingleImage(self.theme.loadSprite(path + '.png', sprites=self))

    def teamGrenade(self, team):
        teamId = getTeamId(team)
        if teamId == 'A':
            path = 'blueGrenade.png'
        elif teamId == 'B':
            path = 'redGrenade.png'
        else:
            return self.grenade
        return self.theme.loadSprite(path, sprites=self)

    @cached
    def zoneBackground(self, team):
        if team is None:
            path = 'greyzone.png'
        elif team.id == 'A':
            path = 'bluezone.png'
        elif team.id == 'B':
            path = 'redzone.png'
        else:
            return self.zoneBackground(None)
        return self.theme.loadSprite(path, sprites=self)

    @cached
    def upgradeImage(self, upgradeType):
        assert issubclass(upgradeType, Upgrade)
        path = upgradeType.iconPath
        if path is None:
            path = 'upgrade-unknown.png'
        return self.theme.loadSprite(
            path, colourkey=upgradeType.iconColourKey, sprites=self)

    def blockBackground(self, block):
        bd = block.defn

        def zoneOwnerId(zone):
            if zone.owner and zone.dark:
                return zone.owner.id
            else:
                return '\x00'

        if bd.kind in ('top', 'btm'):
            if bd.zone is None:
                return None
            owners = (zoneOwnerId(block.zone),)
        else:
            if bd.zone1 is None:
                owner1 = None
            else:
                owner1 = zoneOwnerId(block.zone1)
            if bd.zone2 is None:
                owner2 = None
            else:
                owner2 = zoneOwnerId(block.zone2)
            owners = (owner1, owner2)
        result = self._getBlockFromStoreOrBuild(bd.kind, owners)
        return result

    def _getBlockFromStoreOrBuild(self, kind, owners):
        try:
            return self._store['blockBackground', kind, owners]
        except KeyError:
            self._buildBlockBackgrounds()
            return self._store['blockBackground', kind, owners]

    def getFilledBlockBackground(self, block, owner):
        ownerId = owner.id if owner is not None else '\x00'
        if block.defn.kind in ('top', 'btm'):
            if block.defn.zone is None:
                return None
            else:
                owners = (ownerId,)
        else:
            owner1 = ownerId if block.zone1 is not None else None
            owner2 = ownerId if block.zone2 is not None else None
            owners = (owner1, owner2)
        return self._getBlockFromStoreOrBuild(block.defn.kind, owners)

    def _buildBlockBackgrounds(self):
        '''
        Loads and caches zone backgrounds for all combinations of block owners.
        '''

        filename = 'zone.png'
        flags = self.theme.app.displaySettings.getSurfaceFlagsHack()

        zonePics = self.getZonePics(filename)

        def storePic(kind, owners, pic):
            self._store['blockBackground', kind, owners] = pic

        if self.theme.app.displaySettings.paralaxBackgrounds:
            fillColour = (0, 0, 0, 0)
        else:
            fillColour = BLOCK_BACKGROUND_COLOURKEY

        for kind in ('top', 'btm'):
            for ownerId, zonePic in zonePics.iteritems():
                pic = pygame.Surface(BODY_BLOCK_SCREEN_SIZE, flags)
                pic.fill(fillColour)
                pic.blit(zonePic, BLOCK_OFFSETS[kind])
                storePic(kind, (ownerId,), pic)
        for kind in ('fwd', 'bck'):
            for oid1 in (None, '\x00', 'A', 'B'):
                for oid2 in (None, '\x00', 'A', 'B'):
                    pic = pygame.Surface(
                        INTERFACE_BLOCK_SCREEN_SIZE, flags)
                    pic.fill(fillColour)
                    if oid1 is not None:
                        pic.blit(
                            zonePics[oid1], BLOCK_OFFSETS[kind][0])
                    if oid2 is not None:
                        pic.blit(
                            zonePics[oid2], BLOCK_OFFSETS[kind][1])
                    storePic(kind, (oid1, oid2), pic)

    def getZonePics(self, filename):
        return {
            '\x00': self.theme.loadSprite(
                'grey%s' % (filename,), sprites=self),
            'A': self.theme.loadSprite('blue%s' % (filename,), sprites=self),
            'B': self.theme.loadSprite('red%s' % (filename,), sprites=self),
        }

    @cached
    def orb(self, team):
        teamId = getTeamId(team)
        if teamId == 'A':
            path = 'blueOrb.png'
        elif teamId == 'B':
            path = 'redOrb.png'
        else:
            return self.neutralOrb
        return self.theme.loadSprite(path, sprites=self)

    @cached
    def netOrb(self):
        return self.theme.loadSprite('netOrb.png', sprites=self)

    @cached
    def shotAnimation(self, team):
        teamId = getTeamId(team)

        if teamId == 'A':
            colour = 'blue'
        elif teamId == 'B':
            colour = 'red'
        elif teamId == '\x00':
            colour = 'default'
        else:
            colour = 'team' + teamId
        return self.theme.makeShot(colour, self)


class ThemeSprites(ThemeSpritesBase):
    def __init__(self, theme):
        super(ThemeSprites, self).__init__(theme)
        self.greyed = GreyedSprites(self)

    def doTransform(self, img):
        return img

    def clearCache(self):
        super(ThemeSprites, self).clearCache()
        self.greyed.clearCache()


class GreyedSprites(ThemeSpritesBase):
    def __init__(self, sprites):
        super(GreyedSprites, self).__init__(sprites.theme)

    def doTransform(self, img):
        grey_colour = (100, 100, 100)
        result = pygame.Surface(img.get_size())
        result.fill(grey_colour)
        setAlpha(img, 32)
        result.blit(img, (0, 0))
        result.set_colorkey(grey_colour)
        return result


class Theme(object):
    def __init__(self, app):
        self.app = app
        self.paths = []
        self.colours = ThemeColours()
        self.sprites = ThemeSprites(self)
        self.setTheme("default")
#        self.setTheme(app.displaySettings.theme)

        # Building the different zone backgrounds takes time, so do it once on
        # startup.
        self.sprites._buildBlockBackgrounds()

        app.displaySettings.onDetailLevelChanged.addListener(
            self.detailChanged)

    def detailChanged(self):
        self.sprites.clearCache()
        self.sprites._buildBlockBackgrounds()

    def setTheme(self, themeName):
        '''
        Sets the theme to the theme with the given name.
        '''
        self.name = themeName
        self.paths = [data.getPath(data.user), data.getPath(data)]

        def insertPath(p):
            if os.path.exists(p):
                self.paths.insert(0, p)
        insertPath(data.getPath(data.themes, themeName))
        insertPath(data.getPath(data.user, 'themes', themeName))
        self.initFonts()
        self.initSounds()
        self.initColours()

    def initColours(self):
        colourPath = self.getPath('config', 'colours.cfg')
        colourData = self._getColourData(colourPath)
        defaultColours = self._getColourData(data.getPath(data))

        for colourName, colour in defaultColours.iteritems():
            if colourName in colourData:
                colour = colourData[colourName]
            setattr(self.colours, colourName, colour)

    def initSounds(self):
        self.app.soundPlayer.addSound('buyUpgrade.ogg', 'buyUpgrade', 1)
        self.app.soundPlayer.addSound('gameLose.ogg', 'gameLose', 1)
        self.app.soundPlayer.addSound('startGame.ogg', 'startGame', 1)
        self.app.soundPlayer.addSound('shoot.ogg', 'shoot', 3)
        self.app.soundPlayer.addSound('turret.ogg', 'turret', 1)
        self.app.soundPlayer.addSound(
            'explodeGrenade.ogg', 'explodeGrenade', 1)
        self.app.soundPlayer.addSound('goal.ogg', 'goal', 1)
        self.app.soundPlayer.addSound('ownGoal.ogg', 'ownGoal', 1)

    def initFonts(self):
        fontData = self._getFontData()

        for fontName, defaultDetails in DEFAULT_FONTS.iteritems():
            if fontName in fontData:
                fontFile, size, bold = fontData[fontName]
            else:
                fontFile, size, bold = defaultDetails.unpack()

            if fontName in UNSCALED_FONTS:
                font = Font(fontFile, size, bold)
            else:
                font = ScaledFont(fontFile, size, bold)
            self.app.fonts.addFont(fontName, font)

    def _getFontData(self):
        try:
            fontLines = open(self.getPath('config', 'fonts.cfg')).readlines()
        except IOError:
            return {}

        result = {}
        for line in fontLines:
            bits = line.split("=")
            bits[2] = bits[2].strip()
            # Perform basic checks
            if len(bits) != 3 or not bits[2].isdigit():
                log.warning('Invalid font config line: %r', line)
            else:
                result[bits[0]] = (bits[1], int(bits[2]))
        return result

    def _getColourData(self, filepath):
        try:
            lines = open(self.getPath('config', 'colours.cfg')).readlines()
        except IOError:
            return {}

        result = {}
        for line in lines:
            line = line.strip()
            if line == '' or line.startswith('#'):
                continue
            bits = line.split("=", 1)
            # Perform basic checks
            invalid = False
            if len(bits) != 2:
                invalid = True
            else:
                try:
                    colour = unrepr(bits[1])
                    if type(colour) is str:
                        colour = colour.strip("'")
                except:
                    invalid = True
                else:
                    if colour in result.keys():
                        colour = result[colour]
                    else:
                        if (not isinstance(colour, tuple) or len(colour) < 3 or
                                len(colour) > 4):
                            invalid = True
            if invalid:
                log.warning('Invalid colour config line: %r', line)
            else:
                result[bits[0].strip()] = colour
        return result

    def getPath(self, *pathBits):
        '''
        Returns a path to the given themed file, looking in the following
        locations:
         1. User theme files for the current theme.
         2. Built-in theme files for the current theme.
         3. Default files.
        '''
        for path in self.paths:
            path = os.path.join(path, *pathBits)
            if os.path.isfile(path):
                return path
        raise IOError('file not found: %s' % (os.path.join(*pathBits),))

    def loadSprite(
            self, filename, colourkey=None, sprites=None,
            borders=(127, 127, 127)):
        '''
        Loads the sprite with the given name. A colour key of None may be given
        to disable colourkey transparency.
        '''
        filepath = self.getPath('sprites', filename)
        image = pygame.image.load(filepath)

        alpha = (colourkey is None)
        if alpha:
            image = image.convert_alpha()

            if not self.app.displaySettings.perPixelAlpha:
                # Create a surface that doesn't have the per-pixel alpha
                result = pygame.Surface(image.get_size())
                result.fill(borders)
                result.blit(image, (0, 0))
                result.set_colorkey(borders)
                image = result
        else:
            image = image.convert()
            image.set_colorkey(colourkey)

        if sprites:
            image = sprites.doTransform(image)

        return image

    def loadSprites(self, filenames, sprites, **kwargs):
        images = []
        for filename in filenames:
            images.append(self.loadSprite(filename, sprites=sprites, **kwargs))
        return images

    def makeShot(self, colour, sprites):
        '''
        Creates the shot sprite / animation. This is here because the Pirate
        theme needs the shot to spin, which it does not normally do.
        '''
        shotImg = self.loadSprite('%sShot.png' % (colour,), sprites=sprites)

        if self.name != 'pirate':
            return SingleImage(shotImg)

        numShotImages = 5
        degree = (360 / numShotImages)

        shots = [
            pygame.transform.rotate(shotImg, degree * i)
            for i in range(0, numShotImages)]
        return Animation(0.07, timeNow, *shots)

    def loadTeamSprite(self, filename, teamId, sprites, **kwargs):
        '''
        teamId must be 'A' or 'B'.
        If teamId is 'A', grabs <filename>.png
        If teamId is 'B', grabs <filename>b.png if it exists, or <filename>.png
            otherwise.
        '''
        if teamId == 'B':
            fullFilename = '%sb.png' % (filename,)
            try:
                filepath = self.getPath('sprites', fullFilename)
                if not os.path.isfile(filepath):
                    fullFilename = '%s.png' % (filename,)
            except IOError:
                fullFilename = '%s.png' % (filename,)
        else:
            fullFilename = '%s.png' % (filename,)

        return self.loadSprite(fullFilename, sprites=sprites, **kwargs)

    def loadTeamSprites(self, filenames, teamId, sprites, **kwargs):
        images = []
        for filename in filenames:
            images.append(self.loadTeamSprite(
                filename, teamId, sprites, **kwargs))
        return images


class F(object):

    def __init__(self, fontFile, size, bold=False):
        self.fontFile = fontFile
        self.size = size
        self.bold = bold

    def unpack(self):
        return (self.fontFile, self.size, self.bold)

DEFAULT_FONTS = {
    'default': F('Junction.ttf', 24),
    'defaultTextBoxFont': F('Junction.ttf', 20),
    'unobtrusivePromptFont': F('Junction.ttf', 28),
    'chatFont': F('Junction.ttf', 25),
    'newChatFont': F('Vera.ttf', 14, True),

    'winMessageFont': F('Junction.ttf', 32),

    'nameFont': F('Junction.ttf', 20),
    'countdownFont': F('Junction.ttf', 16),

    'hugeMenuFont': F('Junction.ttf', 54),
    'bigMenuFont': F('Junction.ttf', 36),
    'mainMenuFont': F('Junction.ttf', 36),
    'serverListFont': F('Junction.ttf', 24),
    'timerFont': F('Junction.ttf', 32),
    'consoleFont': F('orbitron-light.ttf', 20),
    'ampleMenuFont': F('Junction.ttf', 40),
    'mediumMenuFont': F('Junction.ttf', 36),
    'menuFont': F('Junction.ttf', 30),
    'smallMenuFont': F('Junction.ttf', 20),
    'ingameMenuFont': F('FreeSans.ttf', 12),
    'gameInfoFont': F('FreeSans.ttf', 14),
    'gameInfoTitleFont': F('FreeSans.ttf', 20),
    'versionFont': F('Junction.ttf', 16),
    'scrollingButtonsFont': F('Junction.ttf', 24),
    'zoneBarFont': F('Junction.ttf', 24),
    'dialogButtonFont': F('KLEPTOCR.TTF', 50),
    'serverSelectionCheckboxesFont': F('Junction.ttf', 28),

    'messageFont': F('Junction.ttf', 20),
    'leaderboardFont': F('FreeSans.ttf', 14),

    'smallNoteFont': F('Junction.ttf', 22),
    'labelFont': F('Junction.ttf', 32),
    'captionFont': F('Junction.ttf', 35),
    'keymapFont': F('Junction.ttf', 20),
    'keymapInputFont': F('Junction.ttf', 20),

    'achievementTitleFont': F('orbitron-light.ttf', 21),
    'achievementNameFont': F('Junction.ttf', 18),

    'connectionFailedFont': F('Junction.ttf', 32),

    'creditsFont': F('Junction.ttf', 24),
    'creditsH2': F('KLEPTOCR.TTF', 48),
    'creditsH1': F('KLEPTOCR.TTF', 60),

    'bigZoneFont': F('VeraBd.ttf', 164),
}

UNSCALED_FONTS = set([
    'nameFont',
    'ingameMenuFont',
    'gameInfoFont',
    'gameInfoTitleFont',
    'leaderboardFont',
    'newChatFont',
    'winMessageFont',
    'bigZoneFont',
])



import logging

from twisted.internet import defer
from twisted.web.client import getPage

from trosnoth.gui.framework import framework
from trosnoth.gui.framework.elements import TextElement
from trosnoth.trosnothgui.settings.settings import SettingsMenu
from trosnoth.trosnothgui.pregame.savedGameMenu import SavedGameMenu
from trosnoth.trosnothgui.pregame.practise import PractiseScreen
from trosnoth.trosnothgui.pregame.playscreen import PlayAuthScreen
from trosnoth.trosnothgui.pregame.serverSelectionScreen import (
        ServerSelectionScreen)
from trosnoth.trosnothgui.common import mainButton
from trosnoth.trosnothgui.credits import CreditsScreen

from trosnoth.trosnothgui.pregame.firstplaynotify import (
        FirstPlayNotificationBar)
import trosnoth.version

from trosnoth.gui.common import ScaledLocation

log = logging.getLogger(__name__)


class StartupInterface(framework.CompoundElement):
    '''Represents the interface while the game is not connected to a server.'''
    connectScreenFactory = PlayAuthScreen
    serverScreenFactory = ServerSelectionScreen
    practiseScreenFactory = PractiseScreen
    creditsScreenFactory = CreditsScreen

    def __init__(self, app, mainInterface):
        super(StartupInterface, self).__init__(app)
        self.interface = mainInterface

        # Create font.
        self.font = self.app.screenManager.fonts.bigMenuFont

        self.offsets = self.app.screenManager.offsets

        # Create other elements.
        self.buttons = [
            self.button('play',         self.playClicked,        (65, 225),
                    hugeFont=True),
            self.button('servers', self.serverSelectionClicked, (85, 285),
                    smallFont=True),
            self.button('practise',     self.practiseClicked,    (85, 325),
                    smallFont=True),
            self.button('archives',     self.savedGamesClicked,  (65, 420)),
            self.button('settings',     self.settingsClicked,    (65, 490)),
            self.button('credits',      self.creditsClicked,     (65, 560)),
            self.button('exit',         self.exitClicked,        (939, 700),
                    'topright')
        ]
        self.firstTimeNotification = FirstPlayNotificationBar(app)
        self.updateNotification = self._makeUpdateNotificationBar()
        self.elements = self.buttons + [
            self.firstTimeNotification,
            self.updateNotification,
        ]

        if app.identitySettings.firstTime:
            self.firstTimeNotification.show()

        # Create sub-menus.
        self.settingsMenu = SettingsMenu(app, onClose=self.mainMenu,
                onRestart=app.restart)
        self.serverSelectionScreen = ServerSelectionScreen(app,
                onClose=self.mainMenu)
        self.savedGameMenu = None
        self.practiseScreen = self.practiseScreenFactory(app,
                onClose=self.mainMenu,
                onStart=self.interface.connectToGameObject)
        self.creditsScreen = self.creditsScreenFactory(self.app,
                self.app.theme.colours.mainMenuColour, self.mainMenu,
                highlight=self.app.theme.colours.mainMenuHighlight)

        # Which sub-menu is currently active.
        self.currentMenu = None

    @defer.inlineCallbacks
    def start(self):
        try:
            stable = yield getPage('http://trosnoth.org/stable-version.txt')
        except Exception as e:
            log.warning(
                'Failed to check trosnoth.org for stable version: {}'.format(
                    e))
            return
        stable = stable.strip()
        if stable != trosnoth.version.version:
            self.updateNotification.show()

    def _makeUpdateNotificationBar(self):
        from trosnoth.gui.common import Location, Area, ScaledPoint, ScaledSize
        from trosnoth.gui.notify import NotificationBar
        bar = NotificationBar(
            self.app,
            message='This is not the latest stable release. Click for info.',
            url='https://trosnoth.org/download',
            font=self.app.fonts.default,
            area=Area(ScaledPoint(0, 0), ScaledSize(1024, 30), 'topleft'),
            buttonPos=Location(ScaledPoint(1024, 0), 'topright'),
            textPos=Area(ScaledPoint(512, 15), ScaledSize(1024, 30), 'centre'),
        )
        bar.onClick.addListener(bar.hide)
        return bar

    def button(self, text, onClick, pos, anchor='topleft', hugeFont=False,
            smallFont=False):
        return mainButton(self.app, text, onClick, pos, anchor, hugeFont,
                smallFont)

    def heading(self, caption):
        return TextElement(self.app, caption, self.font,
                ScaledLocation(1000, 60, 'topright'),
                self.app.theme.colours.headingColour)

    def practiseClicked(self):
        #self.elements = [self.practiseScreen]
        self.practiseScreen.startGame()

    def playClicked(self):
        playAuthScreen = self.connectScreenFactory(self.app,
                onSucceed=self.mainMenu, onFail=self.mainMenu)
        self.elements = [playAuthScreen]

        servers = ['self']
        if self.app.connectionSettings.lanGames == 'beforeinet':
            servers.append('lan')
        servers += list(self.app.connectionSettings.servers)
        if self.app.connectionSettings.otherGames:
            servers.append('others')
        if self.app.connectionSettings.lanGames == 'afterinet':
            servers.append('lan')
        if self.app.connectionSettings.createGames:
            servers.append('create')
        playAuthScreen.begin(tuple(servers))

    def serverSelectionClicked(self):
        self.serverSelectionScreen.reload()
        self.elements = [self.serverSelectionScreen]

    def creditsClicked(self):
        self.creditsScreen.restart()
        self.elements = [self.creditsScreen]

    def exitClicked(self):
        # Quit the game.
        self.app.stop()

    def mainMenu(self):
        self.elements = self.buttons

    def settingsClicked(self):
        self.elements = [self.settingsMenu]

    def savedGamesClicked(self):
        if self.savedGameMenu is None:
            self.savedGameMenu = SavedGameMenu(self.app,
                    onCancel=self.mainMenu, onReplay=self.replayConnect)
        self.savedGameMenu.replayTab.populateList()
        self.elements = [self.savedGameMenu]

    def inetConnect(self, server, port):
        'Called when user selects connect from inet menu.'
        self.mainMenu()

        # Actually connect.
        self.interface.connectToServer(server, port)

    def replayConnect(self, fname):
        self.interface.connectToReplay(fname)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('trosnoth', '0005_auto_20160208_2340'),
    ]

    operations = [
        migrations.CreateModel(
            name='UpgradesUsedInGameRecord',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('upgrade', models.CharField(max_length=1)),
                ('count', models.IntegerField(default=0)),
                ('gamePlayer', models.ForeignKey(to='trosnoth.GamePlayer')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='upgradesusedingamerecord',
            unique_together=set([('gamePlayer', 'upgrade')]),
        ),
        migrations.RemoveField(
            model_name='gameplayer',
            name='upgradesUsed',
        ),
        migrations.AddField(
            model_name='gameplayer',
            name='botName',
            field=models.TextField(default=b'', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='gamerecord',
            name='replayName',
            field=models.TextField(default=b'', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='gamerecord',
            name='winningTeam',
            field=models.CharField(max_length=1, blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='gameplayer',
            name='team',
            field=models.CharField(max_length=1, blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='gameplayer',
            name='user',
            field=models.ForeignKey(to='trosnoth.TrosnothUser', null=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='playerkills',
            name='killee',
            field=models.ForeignKey(related_name='+', to='trosnoth.GamePlayer'),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='playerkills',
            name='killer',
            field=models.ForeignKey(related_name='+', to='trosnoth.GamePlayer', null=True),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='playerkills',
            unique_together=set([('killer', 'killee')]),
        ),
        migrations.RemoveField(
            model_name='playerkills',
            name='game',
        ),
    ]
